//*********************************************************************************
//	SLEPC EIGENBASES v1.0
//
//	THIS PROGRAM UTILIZES THE PETSC AND SLEPC LIBRARY TO GENERATE A SET OF
//	EIGENFUNCTIONS TO THE HAMILTONIAN OPERATOR BASED ON A NONRELATIVISTIC
//	POTENTIAL FUNCTION TO BE SPECIFIED BY PARAMETERS AT COMPILE TIME.
//
//	NOTE THAT WE USE THE C VERSION OF PETSC COMPILED WITH THE FOLLOWING
//	OPTIONS
//	./configure PETSC_ARCH=linux-gnu-intel --with-cc=icc --with-fc=ifort 
//	--with-scalar-type=complex --with-64-bit-indices --with-clanguage=c 
//	--with-mpi=1 --with-x=0 --with-shared-libraries --with-blas-lapack-dir=/opt/intel/composerxe/mkl/lib/intel64/ 
//	--with-debugging=0 --with-mpi-dir=/usr/mpi/intel/openmpi-1.5.4/
//
//	THIS MPI CAPABLE PROGRAM ACCOMPANIES THE MASTER EQUATION CODE SOLVING
//	THE DISSIPATIVE QUARKONIUM EVOLUTION BASED ON 
//
//  ``Heavy quark master equations in the Lindblad form at high temperatures,''
//  Phys. Rev. D 91, no. 5, 056002 (2015)
//  doi:10.1103/PhysRevD.91.056002
//  [arXiv:1403.5783 [hep-ph]].
//
//	AUTHOR: Dr. Alexander Karl Rothkopf	DATE: February 9th 2020
//
//*********************************************************************************

//*********************************************************************************
//
//     Copyright (C) 2020  Dr. Alexander Rothkopf
//
//     This program is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 2 of the License, or
//     (at your option) any later version.
//
//     This program is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License along
//    with this program; if not, write to the Free Software Foundation, Inc.,
//    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//    
//    A copy of the GNU General Public License is also available at
//    http://www.fsf.org/licensing/licenses
//
//*********************************************************************************

//*********************************************************************************
//	STANDARD INCLUDES
//*********************************************************************************

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#define PI (3.141592653589793)

//*********************************************************************************
//    SLEPC INCLUDES
//*********************************************************************************


#undef __FUNCT__
#define __FUNCT__ "main"

//*********************************************************************************
//	IF MPI CAPABILITIES ARE REQUIRED SET SWITCH HERE 
//*********************************************************************************

#define USEMPI 1

#if (USEMPI==1) 
	#include <mpi.h>
#endif

#include "./slepc/private/epsimpl.h"
#include "slepceps.h"

//*****************************************************************************
//   PARAMETERS TO SELECT THE BASIS STATES DEFINING THE OCCUPATION NUMBERS
//*****************************************************************************

    struct StatesParms
    {

    double alpha;
    double sigma;
    double c;   
    double DistReg;
    double mD;
        
    char BasisFuncsHD[256];
    
    };


//*********************************************************************************
//    MAIN PROGRAM ROUTINE STARTS HERE
//*********************************************************************************

int main(int argc,char *argv[])
{
        PetscErrorCode ierr;
    
	//*********************************************************************************
	//    GEOMETRY AND BASIS STATE PARAMETERS
	//*********************************************************************************
    
		

        int BndCnd=0;
        

        struct StatesParms MySParms;
       
		//Takahiro Miura Parameter Set T=0.1

			int64_t BoxSize=256;
			double dx=1.;

			// // int64_t BoxSize=512;
			// // double dx=0.5;

			// int64_t BoxSize=1024;
			// double dx=0.25;

			int WhichPot=3;
			double Mass=1.;
			
			double DistReg=1./Mass;
			
			MySParms.alpha=0.3;
			MySParms.sigma=0.;
			MySParms.c=0.;
			MySParms.mD=0.2;
			MySParms.DistReg=DistReg;

		// //Takahiro Miura Parameter Set T=0.2

		// 	int64_t BoxSize=256;
		// 	double dx=1.;

		// 	int WhichPot=3;
		// 	double Mass=1.;
			
		// 	double DistReg=1./Mass;
			
		// 	MySParms.alpha=0.3;
		// 	MySParms.sigma=0.;
		// 	MySParms.c=0.;
		// 	MySParms.mD=0.4;
		// 	MySParms.DistReg=DistReg;

		//Takahiro Miura Parameter Set T=0.05

			// int64_t BoxSize=256;
			// double dx=1.;

			// int WhichPot=3;
			// double Mass=1.;
			
			// double DistReg=1./Mass;
			
			// MySParms.alpha=0.3;
			// MySParms.sigma=0.;
			// MySParms.c=0.;
			// MySParms.mD=0.1;
			// MySParms.DistReg=DistReg;



// 		//Takahiro Miura Parameter Set T=0.3

// //			int64_t BoxSize=256;
// //			int64_t BoxSize=512;
// 			int64_t BoxSize=1024;
// 			double dx=1.;

// 			int WhichPot=3;
// 			double Mass=1.;
			
// 			double DistReg=1./Mass;
			
// 			MySParms.alpha=0.3;
// 			MySParms.sigma=0.;
// 			MySParms.c=0.;
// 			MySParms.mD=0.6;
// 			MySParms.DistReg=DistReg;

		//Shiori Kajimoto Parameter Set

			// int64_t BoxSize=256;
			// double dx=1.;

			// int WhichPot=0;
			// double Mass=0.48; //0.48731;
			
			// double DistReg=1./Mass;
		
			// MySParms.alpha=0.3;
			// MySParms.sigma=0.;
			// MySParms.c=0.;
			// MySParms.mD=0.04; //0.0406091;
			// MySParms.DistReg=DistReg;


        int64_t WhatStates=1;
        int64_t HowManyStates=10;
           
    //*********************************************************************************
	//    SETUP MPI
	//*********************************************************************************
	
	    int64_t MPISize=1;
	    int64_t MPIRank=0;	
	    
	    int noargs=1;
	    
	    SlepcInitialize(&noargs,&argv,(char*)0,(char*)0);

	    #if (USEMPI==1)
		    int MPISizeTMP,MPIRankTMP;
		    MPI_Status MPISTAT;
		    
		    MPI_Comm_size(PETSC_COMM_WORLD,&MPISizeTMP);
		    MPI_Comm_rank(PETSC_COMM_WORLD,&MPIRankTMP);
		    
		    MPISize=MPISizeTMP;
		    MPIRank=MPIRankTMP;
		    
		    ierr = PetscPrintf(PETSC_COMM_WORLD,"MPI Env: Number of Tasks: %d Current Rank:%d\n",MPISize,MPIRank);CHKERRQ(ierr);
	    #endif		
	
	//*****************************************//
	//  Setup GEOMETRY FOR THE SIMULATION      //
	//*****************************************//
	
	    int64_t NX=BoxSize;
	    int64_t VOLUME=NX;
	    
	    int NDIM=1;
	    
	    double DX=dx;
		
	//*****************************************//
	//  Setup EIGENSOLVER PARAMETERS           //
	//*****************************************//
	
	    int admixnum=HowManyStates;
	    
	    double Tol=1e-17;
	    double maxiter=1e8;

	//*****************************************//
	//  Output parameters to HD                //
	//*****************************************//
	
	    if(MPIRank==0){		
		system("mkdir ./DATA1D");
		
		FILE* PARAM;
		PARAM=fopen("./DATA1D/ParametersEBASIS.dat","w");
		
		fprintf(PARAM, "DX: %e\n", DX);
		fprintf(PARAM, "NX: %ld\n", NX);
		fprintf(PARAM, "How many Eigenfunctions: %d\n", admixnum);
		fclose(PARAM);

		
		fprintf(stderr, "\n*********************************************************************************************\n");
		fprintf(stderr, "***                                                                                       ***\n");
		fprintf(stderr, "***      DETERMINATION OF THE EIGENBASIS OF THE 1D HAMILTONIAN OPERATOR FOR USE IN THE    ***\n");
		fprintf(stderr, "***      STOCHASTIC POTENTIAL SIMULATION                                                  ***\n");
		fprintf(stderr, "***                                                                                       ***\n");
		fprintf(stderr, "*********************************************************************************************\n");	


		//*****************************************//
		//  CALCULATE PARTS OF THE EIGENBASIS OF H //
		//*****************************************//
		
		fprintf(stderr, "\n*********************************************************************************************\n");
		fprintf(stderr, "***      CALCUALTE THE EIGENBASIS OF THE INITIAL HAMILTONIAN                              ***\n");
		fprintf(stderr, "***      USING THE SLEPC AND PETSC FRAMEWORK                                              ***\n");
		fprintf(stderr, "*********************************************************************************************\n");
	    }

	    #if (USEMPI==1)
		    MPI_Barrier(PETSC_COMM_WORLD);
	    #endif

	//*****************************************//
	//  INITIALIZE PETSC/SLEPC VARS&ARRAYS     //
	//*****************************************//
		
	    int64_t fullV=VOLUME;
	    fullV*=admixnum*2;
	    
	    Mat            A;               /* operator matrix */
	    EPS            eps;             /* eigenproblem solver context */
	    EPSType  type;
	    EPSType  typeCheck;
	    PetscReal      error, tol, re, im;
	    PetscScalar    kr, ki;
	    PetscInt       N,RowNum, nev, maxit, i, its, nconv,x;

	    Vec EigVec;
	    
	    RowNum=1+2*NDIM;
	    
	    N = VOLUME*VOLUME;
		    
	    ierr = PetscPrintf(PETSC_COMM_WORLD,"\n1-D Hamiltonian Eigenproblem (sparse matrix version), N=%d (%dx%d grid)\n\n",N,VOLUME,VOLUME);CHKERRQ(ierr);

	//*****************************************//
	// CONSTRUCT THE HAMILTONIAN IN COORD BSIS //
	//*****************************************//

	    #if (USEMPI==0)
	    
		//*****************************************//
		// INITIALIZE THE SERIAL SPARSE MATRIX     //
		//*****************************************//
		
		VecCreate(PETSC_COMM_WORLD,&EigVec);
		VecSetSizes(EigVec,PETSC_DECIDE,VOLUME);	
		VecSetType(EigVec, VECSEQ);

		VecSet(EigVec , 0.0+PETSC_i*0.0);
		ierr = VecAssemblyBegin(EigVec);CHKERRQ(ierr);
		ierr = VecAssemblyEnd(EigVec);CHKERRQ(ierr);	
		
		PetscInt *Nonzeros;
		
		Nonzeros = (PetscInt*) malloc( sizeof(PetscInt) *VOLUME);
		
		for(PetscInt i=0; i<VOLUME; i++) Nonzeros[i]=1+2*NDIM;
		
		MatCreateSeqAIJ(PETSC_COMM_WORLD,VOLUME,VOLUME,RowNum,Nonzeros,&A);
		
	    #else
		
		//*****************************************//
		// INITIALIZE THE MPI SPARSE MATRIX        //
		//*****************************************//
			
		ierr = MatCreateAIJ(PETSC_COMM_WORLD,PETSC_DECIDE,PETSC_DECIDE,VOLUME,VOLUME,  2*NDIM+1 , PETSC_NULL , (2*NDIM)+1, PETSC_NULL , &A); CHKERRQ(ierr);
		
		ierr = VecCreateMPI(PETSC_COMM_WORLD,PETSC_DECIDE,VOLUME,&EigVec); CHKERRQ(ierr);
		VecSet(EigVec , 0.0+PETSC_i*0.0);
		ierr = VecAssemblyBegin(EigVec);CHKERRQ(ierr);
		ierr = VecAssemblyEnd(EigVec);CHKERRQ(ierr);
		
	    #endif		

	//*****************************************//
	// SETTING THE MATRIX ENTRIES              //
	//*****************************************//
		
	    for( x=MPIRank*(NX/MPISize); x<(MPIRank+1)*(NX/MPISize); x++){
		
		PetscPrintf(PETSC_COMM_WORLD,"TASK: %ld CONSTRUCTING THE MATRIX: %ldth POSITION ALONG X XAXIS\n",MPIRank,x);
				
			PetscInt pos=    x;    

			PetscInt posxm1= (x-1+NX)%NX;    
			PetscInt posxp1= (x+1)%NX;    
			
			PetscReal dist=DX*sqrt( (-NX/2.+x)*(-NX/2.+x) );
			PetscReal potre=0.;

            
            double Pot=0.;
            
            if(WhichPot==0){
                
                        double regdist=sqrt(dist*dist+MySParms.DistReg*MySParms.DistReg);
                
                        Pot=-MySParms.alpha*exp(-MySParms.mD*regdist)/regdist + MySParms.sigma*regdist + MySParms.c;
                        
                        //fprintf(stderr, "%ld - %e - %e\n",x,dist,Pot);
                        
            } else if (WhichPot==1){
 
                        double tmp= cosh( sqrt( MySParms.sigma * MySParms.alpha )* dist );

                        Pot= -(MySParms.alpha*0.5)*6./pow(tmp,2);

        
            } else if (WhichPot==2){
 
                        double tmp= cosh( sqrt( MySParms.sigma * MySParms.alpha )* dist );

                        Pot= -(MySParms.alpha*0.5)*30./pow(tmp,2);

        
            } else if (WhichPot==3){
 
                        double regdist=sqrt(dist*dist+MySParms.DistReg*MySParms.DistReg);
                
                        Pot=-MySParms.alpha*exp(-MySParms.mD*dist)/regdist + MySParms.sigma*regdist + MySParms.c;

        
            } else {
                
             fprintf(stderr, "Please choose correct potential 0,1\n"); MPI_Barrier(MPI_COMM_WORLD); MPI_Finalize(); exit(-1);
                
            }

                        
			PetscScalar LAPFAC= 1./(Mass*DX*DX);
			PetscScalar diag=2.*LAPFAC + Pot;
			PetscScalar LAPFACM= -1./(Mass*DX*DX);

			MatSetValues(A,1,&pos,1,&pos,&diag,INSERT_VALUES);
			MatSetValues(A,1,&pos,1,&posxp1,&LAPFACM,INSERT_VALUES);
			MatSetValues(A,1,&pos,1,&posxm1,&LAPFACM,INSERT_VALUES);
            
	    }
	    
	    if(MPIRank==0)fprintf(stderr, "\n*********************************************************************************************\n");
	    if(MPIRank==0)fprintf(stderr, "***      MATRIX ASSEMBLY BY PETSC START                                                   ***\n");
	    if(MPIRank==0)fprintf(stderr, "*********************************************************************************************\n");

	    #if (USEMPI==1)
		    MPI_Barrier(PETSC_COMM_WORLD);
	    #endif

	    MatAssemblyBegin(A,MAT_FINAL_ASSEMBLY);
	    MatAssemblyEnd(A,MAT_FINAL_ASSEMBLY);

	    //MatView(A,	PETSC_VIEWER_STDOUT_WORLD);
	    
	    if(MPIRank==0)fprintf(stderr, "\n*********************************************************************************************\n");
	    if(MPIRank==0)fprintf(stderr, "***      MATRIX ASSEMBLY BY PETSC FINISH                                                  ***\n");
	    if(MPIRank==0)fprintf(stderr, "*********************************************************************************************\n");

	    
	    #if (USEMPI==1)
		    MPI_Barrier(PETSC_COMM_WORLD);
	    #endif
	
	//*****************************************//
	// SETTING UP THE EIGENSOLVER              //
	//*****************************************//

	    /* 
	    Create eigensolver context
	    */
	    ierr = EPSCreate(PETSC_COMM_WORLD,&eps);CHKERRQ(ierr);
	    
	    /*
	    Set operators. In this case, it is a standard eigenvalue problem
	    */
	    ierr = EPSSetOperators(eps,A,PETSC_NULL);CHKERRQ(ierr);
	    
	    ierr = EPSSetProblemType(eps,EPS_HEP);CHKERRQ(ierr);

	    ierr = EPSSetDimensions(eps, admixnum , 2*2*admixnum, 2*2*admixnum);CHKERRQ(ierr);

	    
	    ierr = EPSSetTolerances(eps, Tol , maxiter );CHKERRQ(ierr);
	
	//*****************************************//
	// DEFINE WHAT EIGENPAIRS TO SOLVE FOR     //
	//*****************************************//
	
	    ierr = EPSSetWhichEigenpairs(eps,EPS_SMALLEST_REAL);CHKERRQ(ierr);
			    
	    #if (USEMPI==1)
		    MPI_Barrier(PETSC_COMM_WORLD);
	    #endif
		
	//*****************************************//
	// EXECUTE THE EIGENSOLVER                 //
	//*****************************************//
		    
	    if(MPIRank==0)fprintf(stderr, "\n*********************************************************************************************\n");
	    if(MPIRank==0)fprintf(stderr, "***      EIGENSYSTEM SOLVER START                                                         ***\n");
	    if(MPIRank==0)fprintf(stderr, "*********************************************************************************************\n");
		    
	    ierr = EPSSolve(eps);CHKERRQ(ierr);
	    ierr = EPSGetIterationNumber(eps, &its);CHKERRQ(ierr);
	    ierr = PetscPrintf(PETSC_COMM_WORLD," Number of iterations of the method: %d\n",its);CHKERRQ(ierr);
	    /*
	    Optional: Get some information from the solver and display it
	    */
	    ierr = EPSGetType(eps,&typeCheck);CHKERRQ(ierr);
	    ierr = PetscPrintf(PETSC_COMM_WORLD," Solution method: %s\n\n",typeCheck);CHKERRQ(ierr);
	    ierr = EPSGetDimensions(eps,&nev,PETSC_NULL,PETSC_NULL);CHKERRQ(ierr);
	    ierr = PetscPrintf(PETSC_COMM_WORLD," Number of requested eigenvalues: %d\n",nev);CHKERRQ(ierr);
	    ierr = EPSGetTolerances(eps,&tol,&maxit);CHKERRQ(ierr);
	    ierr = PetscPrintf(PETSC_COMM_WORLD," Stopping condition: tol=%.4g, maxit=%d\n",tol,maxit);CHKERRQ(ierr);

	//*****************************************//
	// SHOW THE RESULTS FOR EIGENVALUES        //
	//*****************************************//
	    
	    ierr = EPSGetConverged(eps,&nconv);CHKERRQ(ierr);
	    ierr = PetscPrintf(PETSC_COMM_WORLD," Number of converged eigenpairs: %d\n\n",nconv);CHKERRQ(ierr);
			    

	    FILE *SLEPOUTVALS=0, *SLEPOUTVECS=0;	
			    
	    if(MPIRank==0){			
		    SLEPOUTVALS = fopen ("./DATA1D/SLEPCEVALS.dat", "w");
	    }	
	
	
	    if (nconv>0) {
		    
		    /*
		    Display eigenvalues and relative errors
		    */
		    
		    ierr = PetscPrintf(PETSC_COMM_WORLD,
			    "           k          ||Ax-kx||/||kx||\n"
			    "   ----------------- ------------------\n" );CHKERRQ(ierr);

		    ierr = PetscPrintf(PETSC_COMM_WORLD," Reading out the eigenvalues       \n\n"); CHKERRQ(ierr);
		    
		    for( i=0; i<admixnum; i++ ) {
			    
			    /* 
			    Get converged eigenpairs: i-th eigenvalue is stored in kr (real part) and
			    ki (imaginary part)
			    */
			    
			    ierr = EPSGetEigenpair(eps,i,&kr,&ki,PETSC_NULL,PETSC_NULL);CHKERRQ(ierr);
			    
			    /*
				    Compute the relative error associated to each eigenpair
			    */
			    ierr = EPSComputeRelativeError(eps,i,&error);CHKERRQ(ierr);

			    #ifdef PETSC_USE_COMPLEX
			    re = PetscRealPart(kr);
			    im = PetscImaginaryPart(kr);
			    #else
			    re = kr;
			    im = ki;
			    #endif 
			    
			    double dblre=re;
			    double dblim=im;
			    int icnt=i;
			    
			    if(MPIRank==0){	
				    fprintf(SLEPOUTVALS,"%d %.15e %.15e\n", icnt, dblre,dblim);
			    }

			    if (im!=0.0) {
				    ierr = PetscPrintf(PETSC_COMM_WORLD," %.20e%+.20%e j %12g\n",re,im,error);CHKERRQ(ierr);
			    } else {
				    ierr = PetscPrintf(PETSC_COMM_WORLD,"   %.20e       %12g\n",re,error);CHKERRQ(ierr); 
			    }
		    }
			    
		    ierr = PetscPrintf(PETSC_COMM_WORLD,"\n" );CHKERRQ(ierr);	
	    }	
		
	    #if (USEMPI==1)
	    MPI_Barrier(PETSC_COMM_WORLD);
	    #endif
		    
	    if(MPIRank==0){	
		    fclose(SLEPOUTVALS);
	    }

	
	//*****************************************//
	// OUTPUT THE RESULTS TO HD                //
	//*****************************************//
	
	    double DEVLenSt;
	    
	    PetscScalar *EigVecArray;
	    EigVecArray = (PetscScalar*) malloc( sizeof(PetscScalar)* VOLUME/MPISize);
	    
	    char NAME[256];

	    double *EigVecDbl;
	    EigVecDbl=(double*) malloc( 2*sizeof(double)* VOLUME/MPISize);
	    
	    int k;
	    for( k=0; k<admixnum; k++ ) {

		    ierr = PetscPrintf(PETSC_COMM_WORLD," Writing out Eigenvector %3d \n\n", k); CHKERRQ(ierr);
		    
		    #if (USEMPI==1)
		    MPI_Barrier(PETSC_COMM_WORLD);
		    #endif
		    
		    EPSGetEigenpair(eps,k,&kr,&ki,EigVec,PETSC_NULL);CHKERRQ(ierr);
		    ierr = EPSComputeRelativeError(eps,k,&error);CHKERRQ(ierr);
			    
		    DEVLenSt=0.;
		    
		    ierr = VecNorm(EigVec,NORM_2,&DEVLenSt);CHKERRQ(ierr);
		    ierr = PetscPrintf(PETSC_COMM_WORLD,"Eigenvector %d with Norm: %.20e to EigVal:%.20e %e\n",k,DEVLenSt,kr,ki);CHKERRQ(ierr);

		    VecGetArray(EigVec,&EigVecArray);
		    
		    #if (USEMPI==1)
		    if(MPIRank==0){
		    #endif
		    
			    sprintf(NAME, "./DATA1D/SLEPCEVEC_%d.dat", k);	
			    SLEPOUTVECS = fopen (NAME, "wb");
							    
			    for (x=0; x<VOLUME/MPISize; x++) {
				    EigVecDbl[2*x]=PetscRealPart(EigVecArray[x]);
				    EigVecDbl[2*x+1]=PetscImaginaryPart(EigVecArray[x]);
				    
				    //fprintf(stderr,"%d %e %e\n",x,EigVecDbl[2*x],EigVecDbl[2*x+1]);
				    
			    }
			    
			    fwrite(EigVecDbl, sizeof(double),2*VOLUME/MPISize, SLEPOUTVECS);

			    fclose(SLEPOUTVECS);
			    
		    #if (USEMPI==1)
			    int COMMUNICATE;
			    ierr=MPI_Send( &COMMUNICATE,1 , MPI_INT, 1 ,1, PETSC_COMM_WORLD );
			    
		    }else if(MPIRank>0&&MPIRank<MPISize-1){
			    
			    int COMMUNICATE;
			    ierr=MPI_Recv( &COMMUNICATE, 1 , MPI_INT, MPIRank-1, MPIRank, MPI_COMM_WORLD , &MPISTAT );
			    
			    sprintf(NAME, "./DATA1D/SLEPCEVEC_%d.dat", k);	
			    SLEPOUTVECS = fopen (NAME, "ab");
							    
			    for (x=0; x<VOLUME/MPISize; x++) {
				    EigVecDbl[2*x]=PetscRealPart(EigVecArray[x]);
				    EigVecDbl[2*x+1]=PetscImaginaryPart(EigVecArray[x]);
			    }
			    
			    fwrite(EigVecDbl, sizeof(double),2*VOLUME/MPISize, SLEPOUTVECS);
			    
			    fclose(SLEPOUTVECS);
			    
			    ierr=MPI_Send( &COMMUNICATE,1 , MPI_INT, MPIRank+1 ,MPIRank+1, PETSC_COMM_WORLD );
			    
		    }else if(MPIRank==MPISize-1){

			    int COMMUNICATE;
			    
			    ierr=MPI_Recv( &COMMUNICATE, 1 , MPI_INT, MPIRank-1, MPIRank, MPI_COMM_WORLD , &MPISTAT );
			    
			    sprintf(NAME, "./DATA1D/SLEPCEVEC_%d.dat", k);	
			    SLEPOUTVECS = fopen (NAME, "ab");
							    
			    for (x=0; x<VOLUME/MPISize; x++) {
				    EigVecDbl[2*x]=PetscRealPart(EigVecArray[x]);
				    EigVecDbl[2*x+1]=PetscImaginaryPart(EigVecArray[x]);
			    }
			    
			    fwrite(EigVecDbl, sizeof(double),2*VOLUME/MPISize, SLEPOUTVECS);

			    fclose(SLEPOUTVECS);
				    
		    }
					    
		    MPI_Barrier(PETSC_COMM_WORLD);
		    #endif

		    VecRestoreArray(EigVec,&EigVecArray);

		    
	    }
	  
	//*****************************************//
	// CLEANUP                                 //
	//*****************************************//
	
	ierr = EPSDestroy(&eps);CHKERRQ(ierr);
	ierr = MatDestroy(&A);CHKERRQ(ierr);
	ierr = SlepcFinalize();CHKERRQ(ierr);

}
