//*********************************************************************************
//
//	THIS PROGRAM SERVES AS A CROSSCHECK TO THE 1D IMPLEMENTATION OF THE MASTER 
//  EQUATION FOR HEAVY QUARKONIUM DERIVED BY Y. AKAMATSU IN
//
//  ``Heavy quark master equations in the Lindblad form at high temperatures,''
//  Phys. Rev. D 91, no. 5, 056002 (2015)
//  doi:10.1103/PhysRevD.91.056002
//  [arXiv:1403.5783 [hep-ph]].
//
//  ON THE LEVEL OF THE DENSITY MATRIX WHICH INCLUDES FULL DISSIPATIVE EFFECTS.
//  THE SAME EQUATION OF MOTION HAS BEEN APPROXIMATELY UNRAVELLED VIA QUANTUM
//  STATE DIFFUSION BY T. MIURA IN
//
//  T.~Miura, Y.~Akamatsu, M.~Asakawa and A.~Rothkopf,
//  ``Quantum Brownian motion of a heavy quark pair in the quark-gluon plasma,''
//  arXiv:1908.06293 [nucl-th].
//
//  THIS CODE DEPLOYS THE NAIVE FINITE DIFFERENCE OPERATOR IN ORDER TO DEMONSTRATE
//  THE FAILURE OF TRACE CONSERVATION WITHOUT THE USE OF AN APPROPRIATE 
//  REPARAMETRIZATION NEUTRAL SUMMATION-BY-PARTS OPERATOR.
//
//	CODE AUTHOR: Dr. Alexander Karl Rothkopf	DATE: April 4th 2020
//
//*********************************************************************************

//*********************************************************************************
//
//     Copysight (C) 2020  Dr. Alexander Rothkopf
//
//     This program is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 3 of the License, or
//     (at your option) any later version.
//
//     This program is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//     You should have received a copy of the GNU General Public License along
//     with this program; if not, write to the Free Software Foundation, Inc.,
//     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//    
//     A copy of the GNU General Public License is also available at
//     http://www.fsf.org/licensing/licenses
//
//*********************************************************************************

//*********************************************************************************
//	STANDARD INCLUDES, USING C++ LISTS
//*********************************************************************************

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include <time.h>

#include <list>
#include <iostream>
#include <fstream>
#include <istream>
#include <complex>
#include <cmath>

#define PI (3.141592653589793)

//*********************************************************************************
//    MPI INCLUDE
//*********************************************************************************

#include <mpi.h>

//*********************************************************************************
//    PETSC & SLEPC INCLUDES
//    PETSC IS DEPLOYE TO IMPLEMENT THE SEMI-IMPLICIT SOLVER
//    SLEPC ALLOWS US TO COMPUTE THE EIGENVECTORS OF THE EVOLUTION OPERATOR
//*********************************************************************************

//#include <petscksp.h>

#undef __FUNCT__
#define __FUNCT__ "main"

#include "./slepc/private/epsimpl.h"
#include "slepceps.h"

using namespace std;

//*********************************************************************************
//   BOUNDARY CONVENTIONS : CORRECTLY SET TO VALUE 1
//*********************************************************************************

#define NOISERESTRICTED (1)

//*********************************************************************************
//   FORWARD DECLARATIONS
//*********************************************************************************

     int VERBOSE;

    //*****************************************************************************
    //   PARAMETERS FOR THE INTERACTIONS V and D, e.g. via POTENTIAL
    //*****************************************************************************

    struct InteractionParms
    {
    double alpha;
    double sigma;
    double c;

    double DistReg;
    
    double mD;
    
    double gamma;
    
    char PotProfileHD[256];
    
    };

    //*****************************************************************************
    //   PARAMETERS FOR THE BACKGROUND EVOLUTION, I.E. BJORKEN vs. FULL HYDRO
    //*****************************************************************************

    struct BackgroundParms
    {
    
    int WhichBEvol;

    double Tinit;
    double Tfinal;
    
    double lcorr;
    
    char TempProfileHD[256];
    
    double lcorrPropTFac;
    double mDPropTFac;
    double gammaPropTFac;
    
    };

    //*****************************************************************************
    //   PARAMETERS TO SELECT THE BASIS STATES DEFINING THE OCCUPATION NUMBERS
    //*****************************************************************************

    struct StatesParms
    {

    double alpha;
    double sigma;
    double c;   
    double mD;   
    double DistReg;
        
    char BasisFuncsHD[256];
    
    };

    //*****************************************************************************
    //   PARAMETERS TO CHARACTERIZE THE INITIAL DENSITY MATRIX
    //*****************************************************************************

    struct InitCondParms
    {

    double sigmas;
    double sigmaR;
    
    int64_t SelectInitState;
    
    int64_t StartStep;
 
    char LocationOfSavedState[256];
    
    
    };

    //*****************************************************************************
    //   CENTRAL CLASS USED TO HOUSE AND EVOLVE THE DENISTY MATRIX
    //*****************************************************************************

    class DensityMatrix
    {
        public:
        
            //*********************************************************************
            //   ALLOCATE THE MAIN ARRAY AND INITIALIZE THE FT
            //*********************************************************************
            void InitializeGeometry(int64_t NsParm, double dsParm, double dtParm, double MassParm, int64_t ThreadNumParm, int MPINumProcs, int MPIMyRank, int CheckMPIThreadCapable, int WhichDynamics);
            
            //*********************************************************************
            //   SETUP HOW THE FUNCTIONS V, F1, \vec{F2} ARE EVALUATED
            //*********************************************************************            
            void SetupInteractions(int VSel, int DSel, InteractionParms Parms);
            
            //*********************************************************************
            //   SETUP HOW THE TEMPERATURE EVOLUTION PROCEEDS
            //*********************************************************************            
            void SetupBackgroundEvol(int WhichEvol, BackgroundParms Parms);
            
            //*********************************************************************
            //   LOAD THE BASIS STATES FROM HD
            //*********************************************************************            
            void SetupBasisStates(int WhatStates, int NumStates, StatesParms Parms);
            
            //*********************************************************************
            //  INITIALIZE THE DENSITY MATRIX
            //*********************************************************************            
            void InitializeDensityMatrix(int WhichInit, InitCondParms Parms);
            
            //*********************************************************************
            //  RETURN THE DENSITY MATRIX EVALUATED IN OCCUPATION NUMBER BASIS
            //*********************************************************************            
            void DensityMatrixInStateBasis(complex<double> *Coefficients);
            
            //*********************************************************************
            //  RETURN THE QUANTITIES THAT CHARACTERIZE STATE OF THE SYSTEM
            //*********************************************************************            
            double ReturnCurrentTime() {return InternalTimeCnt;};
            double ReturnCurrentTemp(){ return (this->*Temperature)(); };
            double ReturnCurrentCorrelationLength(){ return (this->*CorrelationLength)((this->*Temperature)()); };
            double ReturnCurrentDebyeMass(){ return (this->*DebyeMass)((this->*Temperature)()); };
            double ReturnCurrentGamma(){ return (this->*Gamma)((this->*Temperature)()); };
            double ReturnCurrentReV(double dist){ return (this->*V)(dist); };
            double ReturnCurrentImV(double dist){ return (this->*ImV)(dist); };

            int64_t ReturnStartTimeStep(){ return ICParms.StartStep; };
            
            int64_t ReturnNs(){ return Ns; };
            double Returndt(){ return dt; };
            double Returnds(){ return ds; };
            double ReturnMass(){ return Mass; };
            
            complex<double> ReturnTrace();
            double ReturnTraceAbs();
            complex<double> ReturnEnergy();
            void OutputEigenvalsToHD(int64_t tstep, char* OutputDir);
            
            void OutputStatesToHD_ASCII(char* Filename);
            void OutputRhoToHD(char* Filename, int64_t curtime);
            void ReadInRhoFromHD(char* Filename, int64_t* curtime);
            void OutputInteractionsAndTemp(char* Filename);
            void OutputRhoMidSliceRsToHD_ASCII(char* Filename);
            
            void EvolStepCompleted() { InternalTimeCnt++; };
            
            //*********************************************************************
            //  SOLVER RELATED FUNCTIONS
            //********************************************************************* 
            
            void ApplyFullEvolutionOperatorUsingPETSC(double Stepdt);
            
        private:
                                
            //*********************************************************************
            //  NEEDED QUANTITIES
            //*********************************************************************            
            
            InteractionParms IParms;
            BackgroundParms BParms;
            StatesParms SParms;
            InitCondParms ICParms;
            
            int64_t InternalTimeCnt;
            int64_t ReadInFromHD;
            
            int64_t Ns;
            
            int64_t VOLUMEs;
            int64_t VOLUMEGLOBL;
            
            int64_t ThreadNum;
            MPI_Comm MyCom;
            int64_t NodeNum;
            int64_t NodeRank;
            
            int64_t Dynamics;
            
            double ds;
            double dt;
            
            double Mass;
            
            int64_t HowManyStates;
                       
            double (DensityMatrix::*V)  (double);
            double (DensityMatrix::*InitV)  (double);
            double (DensityMatrix::*ImV)  (double);
            
            double (DensityMatrix::*Sigma) (int64_t x, int64_t y);
            double (DensityMatrix::*SigmaSP) (int64_t x, int64_t y);

            double (DensityMatrix::*Kappax)    (int64_t x, int64_t y);            
            double (DensityMatrix::*Kappay)    (int64_t x, int64_t y);
            
            double (DensityMatrix::*Kappaxx)    (int64_t x, int64_t y);            
            double (DensityMatrix::*Kappayy)    (int64_t x, int64_t y);
            
            double (DensityMatrix::*Kappaxy)    (int64_t x, int64_t y);            
           
            double (DensityMatrix::*Temperature) ();
            double (DensityMatrix::*CorrelationLength) (double);
            double (DensityMatrix::*DebyeMass) (double);
            double (DensityMatrix::*Gamma) (double);
            
            complex<double>* States;

            //*********************************************************************
            //  DISTRIBUTED SOLVER PARTS
            //*********************************************************************            

            Vec rho;                                /* Density Matrix */
            
            PetscErrorCode ierr;
            Vec            xvec,bvec;               /* approx solution, RHS, exact solution */
            Mat            AmatFrw,AmatBkw,Emat,RhoMat;        /* linear system matrix */
            KSP            ksp;                     /* linear solver context */
            PetscInt       AIstart,AIend,RhoIstart,RhoIend,RhoMatIstart,RhoMatIend;
            PetscBool      flg = PETSC_FALSE;
            #if defined(PETSC_USE_LOG)
            PetscLogStage stage;
            #endif            
            
            int64_t xstart;            
            int64_t ystart;
            
            int64_t xend;
            int64_t yend;
        
            int64_t Axstart;           
            int64_t Aystart;
            
            int64_t Axend;            
            int64_t Ayend;
        
            int64_t RhoStart;
            int64_t RhoEnd;

            //*****************************************//
            // SETTING UP THE EIGENSOLVER              //
            //*****************************************//
            
            EPS            eps;           
            double Tol=1e-15;
            double maxiter=1e8;
            PetscReal      error, tol, re, im;
            PetscScalar    kr, ki;
            PetscInt       RowNum, nev, maxit, i, its, nconv;
            EPSType  typeCheck;
            
            //*********************************************************************
            //  ACCESSING THE COMPONENTS
            //*********************************************************************            
             
            int64_t Posy( int64_t y);
            int64_t Posx( int64_t x);
            int64_t PosGlobal( int64_t x, int64_t y);
            
            double Distx( int64_t x);
            double Disty( int64_t y);
             
            //*********************************************************************
            //  DIFFERENT BACKGROUND EVOLUTION CHOICES
            //*********************************************************************            
        
            double ConstantTempProfile();
            double SimpleBjorkenTempProfile();
            
            double ConstantCorrelationLength(double T);
            double InverseTemperatureCorrelationLength(double T);
            
            double ConstantDebyeMass(double T);
            double ProptoTemperatureDebyeMass(double T);
            
            double ConstantGamma(double T);
            double ProptoTemperatureGamma(double T);
            
            //*********************************************************************
            //  DIFFERENT INTERACTION CHOICES
            //*********************************************************************            
            
            double PerturbativeReV(double dist);
            double PerturbativeReVTM(double dist);
            double PoeschlTeller2ReV(double dist);
            double PoeschlTeller5ReV(double dist);

            double InitPerturbativeReV(double dist);
            double InitPerturbativeReVTM(double dist);
            double InitPoeschlTeller2ReV(double dist);
            double InitPoeschlTeller5ReV(double dist);

            double PerturbativeImV(double dist);
        
            double NullFunc(int64_t x,int64_t y);
            
            double PerturbativeSigmaStochPot(int64_t x,int64_t y);
            double PerturbativeSigmaFull(int64_t x,int64_t y);
            double PerturbativeKappaxFull(int64_t x,int64_t y);
            double PerturbativeKappayFull(int64_t x,int64_t y);
            double PerturbativeKappaxyFull(int64_t x,int64_t y);
            double PerturbativeKappaxxFull(int64_t x,int64_t y);
            double PerturbativeKappayyFull(int64_t x,int64_t y);
             
    };

    //*****************************************************************************
    //   Internal properties output
    //*****************************************************************************
 
    void DensityMatrix::OutputEigenvalsToHD(int64_t tstep, char* OutputDir)
    {

            FILE* OutputRho;
            char RhoFname[256];
            
            complex<double> *RhoLocal;
            VecGetArray(rho,&RhoLocal);
        
            int64_t LocCnt=0;
            i=1;

            for(int64_t x=xstart; x<xend; x++){
            for(int64_t y=ystart; y<yend; y++){
                                    
                MatSetValues(RhoMat,1,&x,1,&y, &RhoLocal[ LocCnt ],INSERT_VALUES);
                
                LocCnt++;
            }
            }
            
            VecRestoreArray(rho,&RhoLocal);

            MatAssemblyBegin(RhoMat,MAT_FINAL_ASSEMBLY);
            MatAssemblyEnd(RhoMat,MAT_FINAL_ASSEMBLY);
                        
            ierr = EPSSetOperators(eps,RhoMat,PETSC_NULL);
            ierr = EPSSetProblemType(eps,EPS_NHEP);
            ierr = EPSSetDimensions(eps, Ns , Ns, Ns);
            ierr = EPSSetTolerances(eps, Tol , maxiter );            
            
            ierr = EPSSolve(eps);
             
            if (VERBOSE==1) {
            ierr = EPSGetIterationNumber(eps, &its);
            ierr = PetscPrintf(PETSC_COMM_WORLD," Number of iterations of the method: %d\n",its);
            ierr = EPSGetType(eps,&typeCheck);
            ierr = PetscPrintf(PETSC_COMM_WORLD," Solution method: %s\n\n",typeCheck);
            ierr = EPSGetDimensions(eps,&nev,PETSC_NULL,PETSC_NULL);
            ierr = PetscPrintf(PETSC_COMM_WORLD," Number of requested eigenvalues: %d\n",nev);
            ierr = EPSGetTolerances(eps,&tol,&maxit);
            ierr = PetscPrintf(PETSC_COMM_WORLD," Stopping condition: tol=%.4g, maxit=%d\n",tol,maxit);
            
            //*****************************************//
            // SHOW THE RESULTS FOR EIGENVALUES        //
            //*****************************************//

            ierr = EPSGetConverged(eps,&nconv);
            ierr = PetscPrintf(PETSC_COMM_WORLD," Number of converged eigenpairs: %d\n\n",nconv);
            }
            
            if(NodeRank==0){			
                sprintf(RhoFname, "%sDensityMatrixEvals_%ld.dat",OutputDir,tstep);
                OutputRho=fopen(RhoFname,"w");                    
            }	
            
            for( i=0; i<Ns; i++ ) {
                
                ierr = EPSGetEigenpair(eps,i,&kr,&ki,PETSC_NULL,PETSC_NULL);    
                ierr = EPSComputeRelativeError(eps,i,&error);
                
                #ifdef PETSC_USE_COMPLEX
                re = PetscRealPart(kr);
                im = PetscImaginaryPart(kr);
                #else
                re = kr;
                im = ki;
                #endif 
                
                double dblre=re;
                double dblim=im;
                int icnt=i;
                
                if(NodeRank==0){	
                    fprintf(OutputRho,"%d %.15e %.15e\n", icnt, dblre,dblim);
                }
                
            }

            if(NodeRank==0){			
                fclose(OutputRho);
            }	
              
            MPI_Barrier(PETSC_COMM_WORLD);  
              
    }
    
    
    complex<double> DensityMatrix::ReturnTrace()
    {

        complex<double> Trace;
        Trace=0.;
        
        const complex<double>* RhoLocal;
        VecGetArrayRead(rho,&RhoLocal);
        
        int64_t LocCnt=0;
        
            for(int64_t x=xstart; x<xend; x++){
            for(int64_t y=ystart; y<yend; y++){

                if(x==y){
                
                Trace += RhoLocal[ LocCnt ];
                
                }
                
                LocCnt++;
            }
            }
            
            double measure = ds;
            Trace*=measure;
        
        VecRestoreArrayRead(rho,&RhoLocal);    
            
        return Trace;
                
    }
    
    double DensityMatrix::ReturnTraceAbs()
    {

        double Trace;
        Trace=0.;
        
        const complex<double>* RhoLocal;
        VecGetArrayRead(rho,&RhoLocal);
        
        int64_t LocCnt=0;
        
            for(int64_t x=xstart; x<xend; x++){
            for(int64_t y=ystart; y<yend; y++){

                if(x==y){
                
                Trace += std::abs(RhoLocal[ LocCnt ]);
                
                }
                
                LocCnt++;
            }
            }
            
            double measure = ds;
            Trace*=measure;
        
        VecRestoreArrayRead(rho,&RhoLocal);    
            
        return Trace;
                
    }
    
    complex<double> DensityMatrix::ReturnEnergy()
    {

        complex<double> Energy=0.;
       
        //*********************************************************************************
        //  Computing 0.5(H_x+H_y)\delta(x-y)rho(x,y)
        //*********************************************************************************
                 
            for(int64_t x=Axstart; x<Axend; x++){
            for(int64_t y=Aystart; y<Ayend; y++){

                        if(x==y){
                
                        int64_t pos = PosGlobal( x,y );
                        
                        double dist=Distx(x);
                        
                        int64_t posxp1 = PosGlobal( (x+1)%Ns, y);
                        int64_t posxm1 = PosGlobal( (x-1+Ns)%Ns, y );                         

                        complex<double> diagvalRHS = (2./(ds*ds*Mass) + (this->*V)(dist));
                        
                        MatSetValues(Emat,1,&pos,1,&pos,&diagvalRHS,INSERT_VALUES);
                        
                        complex<double> LapFac =-1./(ds*ds*Mass);
                        
                        MatSetValues(Emat,1,&pos,1,&posxp1,&LapFac,INSERT_VALUES);
                        MatSetValues(Emat,1,&pos,1,&posxm1,&LapFac,INSERT_VALUES);
                                     
 
                        }
                 
            }
            }
        
	    MatAssemblyBegin(Emat,MAT_FINAL_ASSEMBLY);
	    MatAssemblyEnd(Emat,MAT_FINAL_ASSEMBLY);

        MatMult(Emat,rho,bvec);        
        
        VecSum(bvec, &Energy);
        
        return Energy*ds;
        
    }
    
    //*****************************************************************************
    //   Position and distance functions
    //*****************************************************************************
    
    int64_t DensityMatrix::Posx( int64_t y)
    {
        return y;
    }

    int64_t DensityMatrix::Posy( int64_t x)
    {
        return x;
    }

    int64_t DensityMatrix::PosGlobal( int64_t x, int64_t y )
    {
      return     Ns*x+y;
    }
 
    double DensityMatrix::Distx( int64_t x)
    {
      return     ds * sqrt(  pow( x - Ns/2 , 2) );
    }
    
    double DensityMatrix::Disty( int64_t y)
    {
      return     ds * sqrt(  pow( y - Ns/2 , 2) );
    }

 
    //*****************************************************************************
    //   Setup routines
    //*****************************************************************************
    
    void DensityMatrix::InitializeGeometry(int64_t NsParm, double dsParm, double dtParm, double MassParm, int64_t ThreadNumParm, int MPINumProcs, int MPIMyRank, int CheckMPIThreadCapable, int WhichDynamics)
    {
       
        NodeNum=MPINumProcs;
        NodeRank=MPIMyRank;

        Ns=NsParm;        
                 
        ThreadNum=ThreadNumParm;
        
        ReadInFromHD=-1;

        VOLUMEGLOBL=Ns*Ns;
        VOLUMEs=Ns;
        
        ds=dsParm;
        dt=dtParm;
        
        Mass=MassParm;
         
        Dynamics=WhichDynamics;
 
        int64_t indiag;
        if( VOLUMEGLOBL/NodeNum < Ns*Ns ) indiag=20;
        
        fprintf(stderr, "Rank:%ld indiag:%ld\n", NodeRank,indiag);
        
        ierr = MatCreateAIJ(PETSC_COMM_WORLD,PETSC_DECIDE,PETSC_DECIDE,VOLUMEGLOBL,VOLUMEGLOBL,  indiag , PETSC_NULL , 10 , PETSC_NULL , &AmatFrw); 
        ierr = MatSetOption(AmatFrw,MAT_KEEP_NONZERO_PATTERN,PETSC_TRUE);
        
        ierr = MatCreateAIJ(PETSC_COMM_WORLD,PETSC_DECIDE,PETSC_DECIDE,VOLUMEGLOBL,VOLUMEGLOBL,  indiag , PETSC_NULL , 10 , PETSC_NULL , &AmatBkw); 
        ierr = MatSetOption(AmatBkw,MAT_KEEP_NONZERO_PATTERN,PETSC_TRUE);
        
        ierr = MatCreateAIJ(PETSC_COMM_WORLD,PETSC_DECIDE,PETSC_DECIDE,VOLUMEGLOBL,VOLUMEGLOBL,  indiag , PETSC_NULL , 5 , PETSC_NULL , &Emat); 
        ierr = MatSetOption(Emat,MAT_KEEP_NONZERO_PATTERN,PETSC_TRUE);
        
        ierr= MatCreateDense(PETSC_COMM_WORLD,PETSC_DECIDE,PETSC_DECIDE,Ns,Ns,NULL,&RhoMat);
        
        ierr = VecCreateMPI(PETSC_COMM_WORLD,PETSC_DECIDE,VOLUMEGLOBL,&rho); 
        ierr = VecCreateMPI(PETSC_COMM_WORLD,PETSC_DECIDE,VOLUMEGLOBL,&bvec);
        ierr = VecCreateMPI(PETSC_COMM_WORLD,PETSC_DECIDE,VOLUMEGLOBL,&xvec);
        
        ierr = MatGetOwnershipRange(AmatFrw,&AIstart,&AIend);
        ierr = MatGetOwnershipRange(RhoMat,&RhoMatIstart,&RhoMatIend);
        ierr = VecGetOwnershipRange(rho,&RhoIstart,&RhoIend);

        RhoStart=RhoMatIstart;
        RhoEnd=RhoMatIend;
        
        xstart= (RhoIstart/(Ns) )%Ns;
        ystart= (RhoIstart )%Ns;
        
        xend= (RhoIend/(Ns))%Ns; if(xend==0)xend=Ns;
        yend= (RhoIend)%Ns;     if(yend==0)yend=Ns;
        
        Axstart= (AIstart/(Ns) )%Ns;
        Aystart= (AIstart)%Ns;
        
        Axend= (AIend/(Ns))%Ns; if(Axend==0)Axend=Ns;        
        Ayend= (AIend)%Ns; if(Ayend==0)Ayend=Ns;
        
        fprintf(stderr, "Rank:%ld - (%ld-%ld) (%ld) - (%ld) -- (%ld) - (%ld)\n" ,NodeRank,RhoIstart,RhoIend,Axstart,Axend,Aystart,Ayend);
        
        ierr = KSPCreate(PETSC_COMM_WORLD,&ksp);
        ierr = EPSCreate(PETSC_COMM_WORLD,&eps);

        ierr = KSPSetOperators(ksp,AmatBkw,AmatBkw);
        ierr = KSPSetType(ksp, KSPGMRES);
        ierr = KSPSetTolerances(ksp,1.e-14,1.e-50,PETSC_DEFAULT,1000);
        
    }

    void DensityMatrix::SetupBackgroundEvol(int WhichEvol, BackgroundParms Parms)
    {
    
        BParms=Parms;
        
        if(WhichEvol==0){
        
            Temperature = &DensityMatrix::ConstantTempProfile;
            CorrelationLength = &DensityMatrix::ConstantCorrelationLength;
            DebyeMass = &DensityMatrix::ConstantDebyeMass;
            Gamma=&DensityMatrix::ConstantGamma;
            
        } else if (WhichEvol==1){
            
            Temperature = &DensityMatrix::SimpleBjorkenTempProfile;
            CorrelationLength = &DensityMatrix::InverseTemperatureCorrelationLength;
            DebyeMass = &DensityMatrix::ProptoTemperatureDebyeMass;
            Gamma=&DensityMatrix::ProptoTemperatureGamma;
            
        } else {
        
            fprintf(stderr, "Currently only constant temperature profile (0) and naive Bjorken (1) implemented\n");
            MPI_Finalize();
            exit(-1);
            
        }
            
    }

    void DensityMatrix::SetupInteractions(int VSel, int DSel, InteractionParms Parms)
    {
        
        IParms=Parms;
        
        if(VSel==0){
        
            V=&DensityMatrix::PerturbativeReV;
            InitV=&DensityMatrix::InitPerturbativeReV;
            
        } else if (VSel==1){
        
            V=&DensityMatrix::PoeschlTeller2ReV;
            InitV=&DensityMatrix::InitPoeschlTeller2ReV;
            
        } else if (VSel==2){
        
            V=&DensityMatrix::PoeschlTeller5ReV;
            InitV=&DensityMatrix::InitPoeschlTeller5ReV;
            
        } else if (VSel==3){
        
            V=&DensityMatrix::PerturbativeReVTM;
            InitV=&DensityMatrix::InitPerturbativeReVTM;
            
        } else {
        
            fprintf(stderr, "Currently only perturbative V implemented\n");
            exit(-1);
            
        }    
        
        if(DSel==0){
        
            fprintf(stderr, "Currently only perturbative D implemented\n");
            exit(-1);
                                 
        } else if (DSel==2){
        
            ImV             =       &DensityMatrix::PerturbativeImV;
            
            Sigma           =       &DensityMatrix::PerturbativeSigmaFull;
            SigmaSP         =       &DensityMatrix::PerturbativeSigmaStochPot;
            
            Kappax          =       &DensityMatrix::PerturbativeKappaxFull;            
            Kappay          =       &DensityMatrix::PerturbativeKappayFull;
            
            Kappaxx         =       &DensityMatrix::PerturbativeKappaxxFull;            
            Kappayy         =       &DensityMatrix::PerturbativeKappayyFull;

            Kappaxy         =       &DensityMatrix::PerturbativeKappaxyFull;         
                        
        } else {
        
            fprintf(stderr, "Currently only perturbative F implemented\n");
            exit(-1);
            
        }    
        
    }

    void DensityMatrix::SetupBasisStates(int WhatStates, int NumStates, StatesParms Parms)
    {

        SParms=Parms;
        
        HowManyStates=NumStates;
        
        States = new complex<double> [HowManyStates*VOLUMEs];
                
        if(WhatStates==0){
             
            for(int64_t statecnt=0; statecnt<HowManyStates; statecnt++){
            
                double Nsm=0;
                
                for(int64_t x=0; x<Ns; x++){
                        
                        int64_t pos = x;  
                        
                        double dist=Distx(x);
                        
                        States[ statecnt*VOLUMEs + pos ] = complex<double>( exp(- dist*dist * (statecnt+1) ) , 0. );
                            
                        Nsm += ds*pow(States[ statecnt*VOLUMEs + pos ].real(),2);
                        
                }

                for(int64_t i=0; i<VOLUMEs; i++) States[ statecnt*VOLUMEs + i ] /= Nsm;
            }

        }else if (WhatStates==2){
            
            char fname[512];
            double *Evals = new double [2*HowManyStates];
            
            if(NodeRank==0) fprintf(stderr, "******************************************************\n");
            if(NodeRank==0) fprintf(stderr, "***  Basis States from the Hamiltonian             ***\n");
            if(NodeRank==0) fprintf(stderr, "***  Reading SLEPC results from HD   Nx:%3ld        ***\n",VOLUMEs);
 
            int evcnt;

            FILE* SLEPCRES, *SLEPCEVECS;

            sprintf(fname, "%sSLEPCEVALS.dat",SParms.BasisFuncsHD); 		
            SLEPCRES = fopen (fname, "r");	
            if(SLEPCRES==0) { fprintf(stderr, "PROBLEM OPENING %s, EXITING\n",fname); exit(-1); }
             
            
            for(int64_t statecnt=0; statecnt<HowManyStates; statecnt++){
		      
                    sprintf(fname, "%sSLEPCEVEC_%ld.dat",SParms.BasisFuncsHD,statecnt);
            
                    SLEPCEVECS = fopen (fname, "rb");	

                    if(SLEPCRES==0||SLEPCEVECS==0) { fprintf(stderr, "NO EIGENBASIS OF THE HAMILTONIAN AVAILABLE %s, EXITING\n",fname); exit(-1); }
            
                    if(NodeRank==0) fprintf(stderr, "***  Reading in Basis State: %3ld                   ***\n",statecnt);
                  
                    fscanf(SLEPCRES,"%d %lf %lf\n", &evcnt, &Evals[2*statecnt],&Evals[2*statecnt+1]);
                    
                    fseek(SLEPCEVECS,0,SEEK_SET);
                    
                    fread(&States[ statecnt*VOLUMEs ],sizeof(double),2*VOLUMEs,SLEPCEVECS);

                    for(int64_t cnt=0; cnt<VOLUMEs; cnt++) States[ statecnt*VOLUMEs + cnt]/=sqrt(ds);
                    
            }
             
            if(NodeRank==0)  fprintf(stderr, "******************************************************\n\n");
            
        }else{
        
            fprintf(stderr, "Currently only Gaussian (0) or PETSC read-in (2) states implemented\n");
            exit(-1);            
            
        }
 
            if(NodeRank==0) fprintf(stderr, "******************************************************\n");
            if(NodeRank==0) fprintf(stderr, "***  Energies of the Basis states:                 ***\n");
            
            
            for(int64_t statecnt=0; statecnt<HowManyStates; statecnt++){     
                
                complex<double> Energy=0.;
                
                for(int64_t x=0; x<Ns; x++){

                int64_t pos    =   x;
                int64_t posxp1 =   (x+1)%Ns;
                int64_t posxm1 =   (x-1+Ns)%Ns;                
                
                double LapFac=1./(ds*ds);
                
                double dist=Disty(x);
                 
                double Pot=(this->*InitV)(dist);
                 
                Energy+=  -std::conj(States[ statecnt*VOLUMEs + pos ])*( -2.*States[ statecnt*VOLUMEs + pos ] + States[ statecnt*VOLUMEs + posxp1 ]  + States[ statecnt*VOLUMEs + posxm1 ] )*LapFac/Mass + std::conj(States[ statecnt*VOLUMEs + pos ])*Pot*States[ statecnt*VOLUMEs + pos ];
                     
                    
                }

                Energy*=ds;
 
            if(NodeRank==0)  fprintf(stderr, "***  State [%4ld]:  %+e  %+e   ***\n", statecnt,Energy.real(),Energy.imag());
 
            }
            if(NodeRank==0) fprintf(stderr, "******************************************************\n\n");
 
            MPI_Barrier(PETSC_COMM_WORLD);
    }
  
    void DensityMatrix::InitializeDensityMatrix(int WhichInit, InitCondParms Parms)
    {
        
        ICParms=Parms;
       
        InternalTimeCnt=0;
                
        if(WhichInit==0){

            if(NodeRank==0) fprintf(stderr, "******************************************************\n");
            if(NodeRank==0) fprintf(stderr, "***  Initialize Density Matrix via Gaussian in R,s  ***\n");
            if(NodeRank==0) fprintf(stderr, "***                                                 ***\n");
            
            complex<double> *RhoLocal;
            VecGetArray(rho,&RhoLocal);
            
            int64_t LocCnt=0;
            
            for(int64_t x=xstart; x<xend; x++){
            for(int64_t y=ystart; y<yend; y++){

                double R = ( x + y - Ns )/2.;
                double s = ( x - y )/2.;
                
                RhoLocal[ LocCnt ]=exp(-R*R/(ICParms.sigmaR*ICParms.sigmaR))*exp(-s*s/(ICParms.sigmaR*ICParms.sigmaR));
                
                LocCnt++;
            }
            }
            
            VecRestoreArray(rho,&RhoLocal);
            
            VecNormalize(rho,NULL);
            
        } else if(WhichInit==1){
                         
            if(NodeRank==0) fprintf(stderr, "******************************************************\n");
            if(NodeRank==0) fprintf(stderr, "***  Initialize Density Matrix via Basis States    ***\n");
            if(NodeRank==0) fprintf(stderr, "***  using state: %3ld                              ***\n",ICParms.SelectInitState);

            complex<double> *RhoLocal;
            VecGetArray(rho,&RhoLocal);
        
            int64_t LocCnt=0;
            
            for(int64_t x=xstart; x<xend; x++){
            for(int64_t y=ystart; y<yend; y++){

                int64_t posS1=x;
                int64_t posS2=y;
                
                RhoLocal[ LocCnt ] = std::conj(States[ ICParms.SelectInitState*VOLUMEs + posS1 ])*States[ ICParms.SelectInitState*VOLUMEs + posS2 ];
                
                LocCnt++;
            }
            }
            
             VecRestoreArray(rho,&RhoLocal);
            
        }else if (WhichInit==10){

            if(NodeRank==0) fprintf(stderr, "******************************************************\n");
            if(NodeRank==0) fprintf(stderr, "***  Initialize Density Matrix via Saved Config    ***\n");

            
            char fname[512];
            sprintf(fname,"%s/CurrentRho.dat",ICParms.LocationOfSavedState);
            
            ReadInRhoFromHD(fname, &ICParms.StartStep);            
            
            InternalTimeCnt=ICParms.StartStep;
            
            if(NodeRank==0) fprintf(stderr, "***  Starting time step: %8ld                  ***\n",ICParms.StartStep);            
            if(NodeRank==0) fprintf(stderr, "******************************************************\n\n");
            
        }else{
        
            fprintf(stderr, "Currently only Guassian initial density matrix implemented\n");
            exit(-1);            
            
        }
        
       
        
    }
  
  
    void DensityMatrix::DensityMatrixInStateBasis(complex<double> *Coefficients)
    {

        const complex<double>* RhoLocal;
        VecGetArrayRead(rho,&RhoLocal);
        
        for(int64_t scnt1=0; scnt1<HowManyStates; scnt1++){  
            for(int64_t scnt2=0; scnt2<HowManyStates; scnt2++){  
        
            complex<double> MatrixElement = 0.;

            int64_t LocCnt=0;
            
            for(int64_t x=xstart; x<xend; x++){
            for(int64_t y=ystart; y<yend; y++){
                
                int64_t posS1    =   x;
                int64_t posS2    =   y;
                                    
                MatrixElement+= States[ scnt1*VOLUMEs + posS1 ]*RhoLocal[ LocCnt ]*std::conj(States[ scnt2*VOLUMEs + posS2 ]);
                 
                LocCnt++;
            }
            }
            
            double VolElem=pow(ds,2);              
            Coefficients[scnt1*HowManyStates+scnt2]=VolElem*MatrixElement;
           
        }}

        VecRestoreArrayRead(rho,&RhoLocal);
                
        }  
  
    //*****************************************************************************
    //   IMPLEMENTATION OF BACKGROUND CHOICES
    //*****************************************************************************
        
    double DensityMatrix::ConstantTempProfile()
    {
        
        return BParms.Tinit;
        
    }
          
    double DensityMatrix::SimpleBjorkenTempProfile()
    {
        
        return BParms.Tinit*( BParms.Tfinal/ ( InternalTimeCnt*dt + BParms.Tfinal));
        
    }
          
          
    double DensityMatrix::ConstantCorrelationLength(double T)
    {
     
        return BParms.lcorr;
        
    }
           
    double DensityMatrix::InverseTemperatureCorrelationLength(double T)
    {
     
        return BParms.lcorrPropTFac/T;
        
    }

    double DensityMatrix::ConstantDebyeMass(double T)
    {
     
        return IParms.mD;
        
    }
    
    double DensityMatrix::ProptoTemperatureDebyeMass(double T)
    {
     
        return BParms.mDPropTFac*T;
        
    }
    
    double DensityMatrix::ConstantGamma(double T)
    {
     
        return IParms.gamma;
        
    }
    
    double DensityMatrix::ProptoTemperatureGamma(double T)
    {
     
        return BParms.gammaPropTFac*T;
        
    }
    
    //*****************************************************************************
    //   IMPLEMENTATION OF INTERACTION CHOICES
    //*****************************************************************************
    
    double DensityMatrix::PerturbativeReV(double dist)
    {
        
        double T=(this->*Temperature)();
        double DebMass=(this->*DebyeMass)(T);
        
        double regdist=sqrt(dist*dist+IParms.DistReg*IParms.DistReg);

        return -IParms.alpha*exp(-DebMass*regdist)/regdist + IParms.sigma*regdist + IParms.c;
        
    }

    double DensityMatrix::PerturbativeReVTM(double dist)
    {

        // different convention of where to use the regularized distance
        // agrees with convention used in QSD unravelling of the master
        // equation in arXiv:1908.06293
        
        double T=(this->*Temperature)();
        double DebMass=(this->*DebyeMass)(T);
        
        double regdist=sqrt(dist*dist+IParms.DistReg*IParms.DistReg);

        return -IParms.alpha*exp(-DebMass*dist)/regdist + IParms.sigma*regdist + IParms.c;
        
    }

    double DensityMatrix::PoeschlTeller2ReV(double dist)
    {
        double tmp= cosh( sqrt( IParms.sigma * IParms.alpha )* dist );

        return -(IParms.alpha*0.5)*6./pow(tmp,2);
        
    }
     
    double DensityMatrix::PoeschlTeller5ReV(double dist)
    {
        double tmp= cosh( sqrt( IParms.sigma * IParms.alpha )* dist );

        return -(IParms.alpha*0.5)*30./pow(tmp,2);
        
    }

    double DensityMatrix::InitPerturbativeReV(double dist)
    {
            
        double regdist=sqrt(dist*dist+IParms.DistReg*IParms.DistReg);

        return -SParms.alpha*exp(-SParms.mD*regdist)/regdist + SParms.sigma*regdist + SParms.c;
        
    }

    double DensityMatrix::InitPerturbativeReVTM(double dist)
    {
            
        double regdist=sqrt(dist*dist+IParms.DistReg*IParms.DistReg);

        return -SParms.alpha*exp(-SParms.mD*dist)/regdist + SParms.sigma*regdist + SParms.c;
        
    }

    double DensityMatrix::InitPoeschlTeller2ReV(double dist)
    {
        double tmp= cosh( sqrt( SParms.sigma * SParms.alpha )* dist );

        return -(SParms.alpha*0.5)*6./pow(tmp,2);
        
    }
     
    double DensityMatrix::InitPoeschlTeller5ReV(double dist)
    {
        double tmp= cosh( sqrt( SParms.sigma * SParms.alpha )* dist );

        return -(SParms.alpha*0.5)*30./pow(tmp,2);
        
    }
    
    
    double DensityMatrix::PerturbativeImV(double dist)
    {
        
        double T=(this->*Temperature)();
        double CorrLen=(this->*CorrelationLength)(T);
        double CorreLenSquare=CorrLen*CorrLen;
        double gamma=(this->*Gamma)(T);
         
        return gamma*exp( -dist*dist/CorreLenSquare );
        
    }

//===============================================================    
// Null Function  
//===============================================================

    double DensityMatrix::NullFunc(int64_t x,int64_t y)
    {
        return 0.;
    }
   
//===============================================================    
//  Stochastic Potential approximation   
//===============================================================    
   
    double DensityMatrix::PerturbativeSigmaStochPot(int64_t x,int64_t y)
    {
        double SigmaVal=0.;
    
        const double T=(this->*Temperature)();
        const double CorrLen=(this->*CorrelationLength)(T);
        const double CorreLenSquare=CorrLen*CorrLen;
        const double CorreLenQuad=CorreLenSquare*CorreLenSquare;
        const double gamma=(this->*Gamma)(T);
        
        int64_t z0= x - Ns/2 ;
        
        #if(NOISERESTRICTED==1)
        if ( z0 < -Ns/4 ) { z0 += Ns/2; }
        if ( z0 >  Ns/4 ) { z0 -= Ns/2; }
        #endif
                
        double rsquare=(z0*z0);
        
        SigmaVal+= gamma*exp( -rsquare/CorreLenSquare );
        
        SigmaVal-= 2.*gamma;
        
        z0= y - Ns/2;
        
        #if(NOISERESTRICTED==1)
        if( z0 < -Ns/4 ) { z0 += Ns/2; }
        if( z0 >  Ns/4 ) { z0 -= Ns/2; }
        #endif

        rsquare=(z0*z0);
        
        SigmaVal+= gamma*exp( -rsquare/CorreLenSquare );
        
        z0=x+y-Ns;
        double z1 = z0/2.;
        
        #if(NOISERESTRICTED==1)
        if( z1 < -Ns/4 ) { z1 += Ns/2; }
        if( z1 >  Ns/4 ) { z1 -= Ns/2; }
        #endif
        
        rsquare=(z1*z1);
        
        SigmaVal-= 2.*gamma*exp( -rsquare/CorreLenSquare );

        z0=x-y;
        z1 = z0/2.;
        
        #if(NOISERESTRICTED==1)
        if( z1 < -Ns/4 ) { z1 += Ns/2; }
        if( z1 >  Ns/4 ) { z1 -= Ns/2; }
        #endif
        
        rsquare=(z1*z1);

        SigmaVal+= 2.*gamma*exp( -rsquare/CorreLenSquare );
                
        return SigmaVal;
    
    }
    
//===============================================================    
//  Perturbatively derived contributions to the master equation    
//===============================================================    
    
    double DensityMatrix::PerturbativeSigmaFull(int64_t x,int64_t y)
    {
        double SigmaVal=0.;
    
        const double T=(this->*Temperature)();
        const double CorrLen=(this->*CorrelationLength)(T);
        const double CorreLenSquare=CorrLen*CorrLen;
        const double CorreLenQuad=CorreLenSquare*CorreLenSquare;
        const double CorreLenOct=CorreLenQuad*CorreLenQuad;
        
        const double gamma=(this->*Gamma)(T);
        const double denom=16.*Mass*Mass*T*T*CorreLenOct;
        
        const double denomF=8.*Mass*Mass*T*T*CorreLenQuad;

            int64_t z0= x - Ns/2;

            #if(NOISERESTRICTED==1)
            if ( z0 < -Ns/4 ) { z0 += Ns/2; }
            if ( z0 >  Ns/4 ) { z0 -= Ns/2; }
            #endif

            double rsquare=(z0*z0);
        
        // +F_1(x)
        
        SigmaVal+= gamma*exp( -rsquare/CorreLenSquare )*( 4.* rsquare*rsquare +3.*CorreLenQuad + 8.*Mass*T*CorreLenQuad*CorreLenSquare*(-1.+2.*Mass*T*CorreLenSquare) + 4.*rsquare*CorreLenSquare*(-3.+4.*Mass*T*CorreLenSquare) )/denom;
        
            z0= y - Ns/2;

            #if(NOISERESTRICTED==1)
            if ( z0 < -Ns/4 ) { z0 += Ns/2; }
            if ( z0 >  Ns/4 ) { z0 -= Ns/2; }
            #endif

            rsquare=(z0*z0);
        
        // +F_1(y)
        
        SigmaVal+= gamma*exp( -rsquare/CorreLenSquare )*( 4.* rsquare*rsquare +3.*CorreLenQuad + 8.*Mass*T*CorreLenQuad*CorreLenSquare*(-1.+2.*Mass*T*CorreLenSquare) + 4.*rsquare*CorreLenSquare*(-3.+4.*Mass*T*CorreLenSquare) )/denom;
        
        // -2 F_1(0)
        
        SigmaVal-= 2.*gamma*( 3.*CorreLenQuad + 8.*Mass*T*CorreLenQuad*CorreLenSquare*(-1.+2.*Mass*T*CorreLenSquare) )/denom;
 
            z0=x-y;
            double z1 = z0/2.;
            
            #if(NOISERESTRICTED==1)
            if( z1 < -Ns/4 ) { z1 += Ns/2; }
            if( z1 >  Ns/4 ) { z1 -= Ns/2; }
            #endif

            rsquare=(z1*z1);

        // +2*F_1( (x-y)/2 )
        SigmaVal+= 2.*gamma*exp( -rsquare/CorreLenSquare )*( 4.*rsquare*rsquare +3.*CorreLenQuad + 8.*Mass*T*CorreLenQuad*CorreLenSquare*(-1.+2.*Mass*T*CorreLenSquare) + 4.*rsquare*CorreLenSquare*(-3.+4.*Mass*T*CorreLenSquare) )/denom;

            z0=x+y-Ns;
            z1 = z0/2.;
            
            #if(NOISERESTRICTED==1)
            if( z1 < -Ns/4 ) { z1 += Ns/2; }
            if( z1 >  Ns/4 ) { z1 -= Ns/2; }
            #endif

            rsquare=(z1*z1);
    
        // -2F_1( (x+y)/2 )
        SigmaVal-= 2.*gamma*exp( -rsquare/CorreLenSquare )*( 4.* rsquare*rsquare +3.*CorreLenQuad + 8.*Mass*T*CorreLenQuad*CorreLenSquare*(-1.+2.*Mass*T*CorreLenSquare) + 4.*rsquare*CorreLenSquare*(-3.+4.*Mass*T*CorreLenSquare) )/denom;

        //Now come the discretized A terms
        
                z0= x - Ns/2;
                int64_t z2= (x + 2)%Ns - Ns/2;
                int64_t z3= (x - 2 + Ns)%Ns - Ns/2;
                
                #if(NOISERESTRICTED==1)
                if ( z0 < -Ns/4 ) { z0 += Ns/2; }
                if ( z0 >  Ns/4 ) { z0 -= Ns/2; }

                if( z2 < -Ns/4 ) { z2 += Ns/2; }
                if( z2 >  Ns/4 ) { z2 -= Ns/2; }

                if( z3 < -Ns/4 ) { z3 += Ns/2; }
                if( z3 >  Ns/4 ) { z3 -= Ns/2; }
                #endif
        
                rsquare=(z0*z0);
                double rsquare2=(z2*z2);
                double rsquare3=(z3*z3);
        

        SigmaVal-= 0.5*(-1.)*(1./(4.*ds*ds)) *(gamma/denomF) *( 
                +      exp( -rsquare3/CorreLenSquare )*( -2.*rsquare3 + CorreLenSquare)
                -   2.*exp( -rsquare /CorreLenSquare )*( -2.*rsquare  + CorreLenSquare)
                +      exp( -rsquare2/CorreLenSquare )*( -2.*rsquare2 + CorreLenSquare) );
 
        
                z2= (y + 2)%Ns - Ns/2;
                z3= (y - 2 + Ns)%Ns - Ns/2;
        
                #if(NOISERESTRICTED==1)
                if( z2 < -Ns/4 ) { z2 += Ns/2; }
                if( z2 >  Ns/4 ) { z2 -= Ns/2; }

                if( z3 < -Ns/4 ) { z3 += Ns/2; }
                if( z3 >  Ns/4 ) { z3 -= Ns/2; }
                #endif
        
                rsquare2=(z2*z2);
                rsquare3=(z3*z3);
        
        SigmaVal-= 0.5*(-1.)*(1./(4.*ds*ds)) *(gamma/denomF) *( 
                +      exp( -rsquare3/CorreLenSquare )*( -2.*rsquare3 + CorreLenSquare)
                -   2.*exp( -rsquare /CorreLenSquare )*( -2.*rsquare  + CorreLenSquare)
                +      exp( -rsquare2/CorreLenSquare )*( -2.*rsquare2 + CorreLenSquare) );
    
        return SigmaVal;
    
    }

    double DensityMatrix::PerturbativeKappaxFull(int64_t x,int64_t y)
    {
        double KappaVal=0.;
    
        const double T=(this->*Temperature)();
        const double CorrLen=(this->*CorrelationLength)(T);
        const double CorreLenSquare=CorrLen*CorrLen;
        const double CorreLenQuad=CorreLenSquare*CorreLenSquare;
        const double gamma=(this->*Gamma)(T);
        const double denom=8.*Mass*Mass*T*T*CorreLenQuad*CorreLenSquare;            
        const double denomF=8.*Mass*Mass*T*T*CorreLenQuad;

            int64_t z0=x-Ns/2;

            #if(NOISERESTRICTED==1)
            if ( z0 < -Ns/4 ) { z0 += Ns/2; }
            if ( z0 >  Ns/4 ) { z0 -= Ns/2; }
            #endif

            double r=z0;
            double rsquare=(z0*z0);
        
        // +2*F_2(x)
        KappaVal+= 2.*(-1.)*gamma*(r)*exp( -rsquare/CorreLenSquare )*( 2.*rsquare -3.*CorreLenSquare + 4.*T*Mass*CorreLenQuad )/denom;
                
            z0 = x-y;
            double z1 = z0/2.;
            
            #if(NOISERESTRICTED==1)
            if ( z1 < -Ns/4 ) { z1 += Ns/2; }
            if ( z1 >  Ns/4 ) { z1 -= Ns/2; }
            #endif

            r=z1;
            rsquare=(z1*z1);
        
        // +2*F_2( (x-y)/2 )
        KappaVal+= 2.*(-1.)*gamma*(r)*exp( -rsquare/CorreLenSquare )*( 2.*rsquare -3.*CorreLenSquare + 4.*T*Mass*CorreLenQuad )/denom;
        
            z0=x+y-Ns;
            z1 = z0/2.;

            #if(NOISERESTRICTED==1)
            if( z1 < -Ns/4 ) { z1 += Ns/2; }
            if( z1 >  Ns/4 ) { z1 -= Ns/2; }
            #endif
                
            r=z1;
            rsquare=(z1*z1);
        
        //-2F_2( (x+y)/2 ) 
        KappaVal-= 2.*(-1.)*gamma*(r)*exp( -rsquare/CorreLenSquare )*( 2.*rsquare -3.*CorreLenSquare + 4.*T*Mass*CorreLenQuad )/denom;     

        //Now come the discretized A terms

            int64_t z2= (x + 1)%Ns - Ns/2;
            int64_t z3= (x - 1 + Ns)%Ns - Ns/2;

            #if(NOISERESTRICTED==1)
            if( z2 < -Ns/4 ) { z2 += Ns/2; }
            if( z2 >  Ns/4 ) { z2 -= Ns/2; }

            if( z3 < -Ns/4 ) { z3 += Ns/2; }
            if( z3 >  Ns/4 ) { z3 -= Ns/2; }
            #endif
            
            double rsquare2=(z2*z2);
            double rsquare3=(z3*z3);

        KappaVal-=  (1./(2.*ds)) *(-1.)*2.*(gamma/denomF) *( 
                +      exp( -rsquare2/CorreLenSquare )*( -2.*rsquare2 + CorreLenSquare)
                -      exp( -rsquare3/CorreLenSquare )*( -2.*rsquare3 + CorreLenSquare) );        
        
        return KappaVal;
    
    }  
    
    double DensityMatrix::PerturbativeKappayFull(int64_t x,int64_t y)
    {
        double KappaVal=0.;
    
        const double T=(this->*Temperature)();
        const double CorrLen=(this->*CorrelationLength)(T);
        const double CorreLenSquare=CorrLen*CorrLen;
        const double CorreLenQuad=CorreLenSquare*CorreLenSquare;
        const double gamma=(this->*Gamma)(T);
        const double denom=8.*Mass*T*T*CorreLenQuad*CorreLenSquare;    
        const double denomF=8.*Mass*Mass*T*T*CorreLenQuad;
        
            int64_t z0=y-Ns/2;
            
            #if(NOISERESTRICTED==1)
            if ( z0 < -Ns/4 ) { z0 += Ns/2; }
            if ( z0 >  Ns/4 ) { z0 -= Ns/2; }
            #endif

            double r=z0;
            double rsquare=(z0*z0);
        
        // +2*F_2(y)
        KappaVal+= 2.*(-1.)*gamma*(r)*exp( -rsquare/CorreLenSquare )*( 2.*rsquare -3.*CorreLenSquare + 4.*T*Mass*CorreLenQuad )/denom;
                
            z0 = x-y;
            double z1 = z0/2.;

            #if(NOISERESTRICTED==1)
            if( z1 < -Ns/4 ) { z1 += Ns/2; }
            if( z1 >  Ns/4 ) { z1 -= Ns/2; }
            #endif

            r=z1;
            rsquare=(z1*z1);
        
        // -2*F_2( (x-y)/2 )
        KappaVal-= 2.*(-1.)*gamma*(r)*exp( -rsquare/CorreLenSquare )*( 2.*rsquare -3.*CorreLenSquare + 4.*T*Mass*CorreLenQuad )/denom;
        
            z0=x+y-Ns;
            z1 = z0/2.;

            #if(NOISERESTRICTED==1)
            if( z1 < -Ns/4 ) { z1 += Ns/2; }
            if( z1 >  Ns/4 ) { z1 -= Ns/2; }
            #endif

            r=z1;
            rsquare=(z1*z1);
        
        //-2F_2( (x+y)/2 ) 
        KappaVal-= 2.*(-1.)*gamma*(r)*exp( -rsquare/CorreLenSquare )*( 2.*rsquare -3.*CorreLenSquare + 4.*T*Mass*CorreLenQuad )/denom;        

        //Now come the discretized A terms

            int64_t z2= (y + 1)%Ns - Ns/2;
            int64_t z3= (y - 1 + Ns)%Ns - Ns/2;

            #if(NOISERESTRICTED==1)
            if( z2 < -Ns/4 ) { z2 += Ns/2; }
            if( z2 >  Ns/4 ) { z2 -= Ns/2; }

            if( z3 < -Ns/4 ) { z3 += Ns/2; }
            if( z3 >  Ns/4 ) { z3 -= Ns/2; }
            #endif

            double rsquare2=(z2*z2);
            double rsquare3=(z3*z3);

        KappaVal-=  (1./(2.*ds)) *(-1.)*2.*(gamma/denomF) * ( 
                +      exp( -rsquare2/CorreLenSquare )*( -2.*rsquare2 + CorreLenSquare)
                -      exp( -rsquare3/CorreLenSquare )*( -2.*rsquare3 + CorreLenSquare) );
                
        return KappaVal;
   
    }    
    
    
    double DensityMatrix::PerturbativeKappaxyFull(int64_t x,int64_t y)
    {
        double KappaVal=0.;
    
        const double T=(this->*Temperature)();
        const double CorrLen=(this->*CorrelationLength)(T);
        const double CorreLenSquare=CorrLen*CorrLen;
        const double CorreLenQuad=CorreLenSquare*CorreLenSquare;
        const double gamma=(this->*Gamma)(T);
        const double denomF=8.*Mass*Mass*T*T*CorreLenQuad;    
        double z1;
        
            int64_t z0=x-y;
                    z1 = z0/2.;
            
            #if(NOISERESTRICTED==1)
            if( z1 < -Ns/4 ) { z1 += Ns/2; }
            if( z1 >  Ns/4 ) { z1 -= Ns/2; }
            #endif

            double rsquare=(z1*z1);
        
        // +2*F_3( (x-y)/2 )
        
        KappaVal+= 2.*gamma*exp( -rsquare/CorreLenSquare )*( -2.*rsquare + CorreLenSquare )/denomF;
        
            z0=x+y-Ns;
            z1 = z0/2.;
            
            #if(NOISERESTRICTED==1)
            if( z1 < -Ns/4 ) { z1 += Ns/2; }
            if( z1 >  Ns/4 ) { z1 -= Ns/2; }
            #endif

            rsquare=(z1*z1);
        
        // +2*F_3( (x+y)/2 )
           
        KappaVal+= 2.*gamma*exp( -rsquare/CorreLenSquare )*( -2.*rsquare + CorreLenSquare )/denomF;
       
        return KappaVal;
    
    }     
    
    double DensityMatrix::PerturbativeKappaxxFull(int64_t x,int64_t y)
    {
        double KappaVal=0.;
    
        const double T=(this->*Temperature)();
        const double CorrLen=(this->*CorrelationLength)(T);
        const double CorreLenSquare=CorrLen*CorrLen;
        const double CorreLenQuad=CorreLenSquare*CorreLenSquare;
        const double gamma=(this->*Gamma)(T);
        const double denomF=8.*Mass*Mass*T*T*CorreLenQuad;    

            int64_t z0=x-Ns/2;
            
            #if(NOISERESTRICTED==1)
            if ( z0 < -Ns/4 ) { z0 += Ns/2; }
            if ( z0 >  Ns/4 ) { z0 -= Ns/2; }
            #endif

            double rsquare=(z0*z0);
       
        // +F_3(x)

        KappaVal+= gamma*exp( -rsquare/CorreLenSquare )*( -2.*rsquare + CorreLenSquare )/denomF;
        
        // +F_3(0)
 
        KappaVal+= gamma*( CorreLenSquare )/denomF;
        
        return KappaVal;
    
    }    
    
    double DensityMatrix::PerturbativeKappayyFull(int64_t x,int64_t y)
    {
        double KappaVal=0.;
    
        const double T=(this->*Temperature)();
        const double CorrLen=(this->*CorrelationLength)(T);
        const double CorreLenSquare=CorrLen*CorrLen;
        const double CorreLenQuad=CorreLenSquare*CorreLenSquare;
        const double gamma=(this->*Gamma)(T);
        const double denomF=8.*Mass*Mass*T*T*CorreLenQuad;    

            int64_t z0=y-Ns/2;
            
            #if(NOISERESTRICTED==1)
            if ( z0 < -Ns/4 ) { z0 += Ns/2; }
            if ( z0 >  Ns/4 ) { z0 -= Ns/2; }
            #endif

            double rsquare=(z0*z0);
        
        // +F_3(y)
        
        KappaVal+= gamma*exp( -rsquare/CorreLenSquare )*( -2.*rsquare + CorreLenSquare )/denomF;
    
        // +F_3(0)

        KappaVal+= gamma*( CorreLenSquare )/denomF;
        
        return KappaVal;
    
    }    
    
    //*****************************************************************************
    //   IMPLEMENTATION OF THE PETSC SOLVER
    //   1d Cranck-Nicholson based on 1st order Pade rational approximation
    //*****************************************************************************

    void DensityMatrix::ApplyFullEvolutionOperatorUsingPETSC(double Stepdt)
    {
 
        int64_t x,s;

        const double onedivdx=1./ds;
        const double onedivdxdx=1./(ds*ds);     

        const double NewDsecderdsds=1./(16.*ds*ds);
        const double NewDderds=1./(4.*ds);

        //*********************************************************************************
        //  RHS
        //*********************************************************************************
                 
        int64_t LocCnt=0;

        
        if(BParms.WhichBEvol!=0 || InternalTimeCnt==0 || InternalTimeCnt==ReadInFromHD ){

            for(int64_t x=Axstart; x<Axend; x++){
            for(int64_t y=Aystart; y<Ayend; y++){

                        const int64_t pos = PosGlobal( x, y );
                
                        const int64_t posxp1y = PosGlobal( (x+1)%Ns, y );
                        const int64_t posxm1y = PosGlobal( (x-1+Ns)%Ns, y );                         
                        
                        const int64_t posxp2y = PosGlobal( (x+2)%Ns, y );
                        const int64_t posxm2y = PosGlobal( (x-2+Ns)%Ns, y );                         
                        
                        const int64_t posxyp1 = PosGlobal( x, (y+1)%Ns);
                        const int64_t posxym1 = PosGlobal( x, (y-1+Ns)%Ns);                         
                         
                        const int64_t posxyp2 = PosGlobal( x, (y+2)%Ns);
                        const int64_t posxym2 = PosGlobal( x, (y-2+Ns)%Ns);                         
                                         
                        //*****************************************
                        // NEW DERIVATIVE OPERATOR REQUIRES STEPS  
                        // IN BOTH X- AND Y-DIRECTION			   
                        //*****************************************

                        const int64_t posxp1yp1 = PosGlobal( (x+1)%Ns, (y+1)%Ns);
                        const int64_t posxm1ym1 = PosGlobal( (x-1+Ns)%Ns, (y-1+Ns)%Ns);                                                 
                        const int64_t posxp1ym1 = PosGlobal( (x+1)%Ns, (y-1+Ns)%Ns);
                        const int64_t posxm1yp1 = PosGlobal( (x-1+Ns)%Ns, (y+1)%Ns);                         

                        const int64_t posxp2yp2 = PosGlobal( (x+2)%Ns, (y+2)%Ns   );
                        const int64_t posxm2yp2 = PosGlobal( (x-2+Ns)%Ns, (y+2)%Ns);                         
                        const int64_t posxp2ym2 = PosGlobal( (x+2)%Ns, (y-2+Ns)%Ns);
                        const int64_t posxm2ym2 = PosGlobal( (x-2+Ns)%Ns, (y-2+Ns)%Ns);
                       
if(Dynamics==0){
    
                        //*********************************************************************************
                        //  x^2 - contributions - Hamiltonian
                        //*********************************************************************************
                        
                        complex<double> xxderPrefac=  0.5*Stepdt*onedivdxdx*complex<double>(0.,1./Mass);
                        
                        complex<double> xpls1valRHS =      xxderPrefac ;
                        complex<double> xvalRHS     =  -2.*xxderPrefac ;
                        complex<double> xmns1valRHS =      xxderPrefac ;

                        MatSetValues(AmatFrw,1,&pos,1,&posxp1y,&xpls1valRHS,INSERT_VALUES);
                        MatSetValues(AmatFrw,1,&pos,1,&pos   ,&xvalRHS    ,INSERT_VALUES);                        
                        MatSetValues(AmatFrw,1,&pos,1,&posxm1y,&xmns1valRHS,INSERT_VALUES);
 
                        //*********************************************************************************
                        //  y^2 - contributions - Hamiltonian
                        //*********************************************************************************
                        
                        complex<double> yyderPrefac=  0.5*Stepdt*onedivdxdx*complex<double>(0.,-1./Mass);
                        
                        complex<double> ypls1valRHS =      yyderPrefac ;
                        complex<double> yvalRHS     =  -2.*yyderPrefac ;
                        complex<double> ymns1valRHS =      yyderPrefac ;

                        MatSetValues(AmatFrw,1,&pos,1,&posxyp1,&ypls1valRHS,INSERT_VALUES);
                        MatSetValues(AmatFrw,1,&pos,1,&pos   ,&yvalRHS    ,ADD_VALUES);                        
                        MatSetValues(AmatFrw,1,&pos,1,&posxym1,&ymns1valRHS,INSERT_VALUES);
                        
                        //*********************************************************************************
                        //  diagonal terms
                        //*********************************************************************************
                        
                        double distx=Distx(x);
                        double disty=Disty(y);
                        
                        complex<double> diagvalRHS =  complex<double>(1.,0.) + complex<double>( 0. , -Stepdt*0.5 * ( (this->*V)(distx) - (this->*V)(disty) ));
                        
                        MatSetValues(AmatFrw,1,&pos,1,&pos    ,&diagvalRHS,ADD_VALUES);
    
}else if (Dynamics==1){
    
                        //*********************************************************************************
                        //  x^2 - contributions - Hamiltonian
                        //*********************************************************************************
                        
                        complex<double> xxderPrefac=  0.5*Stepdt*onedivdxdx*complex<double>(0.,1./Mass);
                        
                        complex<double> xpls1valRHS =      xxderPrefac ;
                        complex<double> xvalRHS     =  -2.*xxderPrefac ;
                        complex<double> xmns1valRHS =      xxderPrefac ;

                        MatSetValues(AmatFrw,1,&pos,1,&posxp1y,&xpls1valRHS,INSERT_VALUES);
                        MatSetValues(AmatFrw,1,&pos,1,&pos   ,&xvalRHS    ,INSERT_VALUES);                        
                        MatSetValues(AmatFrw,1,&pos,1,&posxm1y,&xmns1valRHS,INSERT_VALUES);
 
                        //*********************************************************************************
                        //  y^2 - contributions - Hamiltonian
                        //*********************************************************************************
                        
                        complex<double> yyderPrefac=  0.5*Stepdt*onedivdxdx*complex<double>(0.,-1./Mass);
                        
                        complex<double> ypls1valRHS =      yyderPrefac ;
                        complex<double> yvalRHS     =  -2.*yyderPrefac ;
                        complex<double> ymns1valRHS =      yyderPrefac ;

                        MatSetValues(AmatFrw,1,&pos,1,&posxyp1,&ypls1valRHS,INSERT_VALUES);
                        MatSetValues(AmatFrw,1,&pos,1,&pos   ,&yvalRHS    ,ADD_VALUES);                        
                        MatSetValues(AmatFrw,1,&pos,1,&posxym1,&ymns1valRHS,INSERT_VALUES);
                        
                        //*********************************************************************************
                        //  diagonal terms
                        //*********************************************************************************
                        
                        double distx=Distx(x);
                        double disty=Disty(y);
                        
                        complex<double> diagvalRHS =  complex<double>(1.,0.) + complex<double>( Stepdt*0.5*(this->*SigmaSP)(x,y) , -Stepdt*0.5 * ( (this->*V)(distx) - (this->*V)(disty) ));
                        
                        MatSetValues(AmatFrw,1,&pos,1,&pos    ,&diagvalRHS,ADD_VALUES);
                        
}else if (Dynamics==2){

                        //*********************************************************************************
                        //  x^2 - contributions - Hamiltonian
                        //*********************************************************************************
                        
                        complex<double> xxderPrefac=  0.5*Stepdt*onedivdxdx*complex<double>(0.,1./Mass);
                        
                        complex<double> xpls1valRHS =      xxderPrefac ;
                        complex<double> xvalRHS     =  -2.*xxderPrefac ;
                        complex<double> xmns1valRHS =      xxderPrefac ;

                        MatSetValues(AmatFrw,1,&pos,1,&posxp1y,&xpls1valRHS,INSERT_VALUES);
                        MatSetValues(AmatFrw,1,&pos,1,&pos    ,&xvalRHS    ,INSERT_VALUES);                        
                        MatSetValues(AmatFrw,1,&pos,1,&posxm1y,&xmns1valRHS,INSERT_VALUES);
 
                        //*********************************************************************************
                        //  y^2 - contributions - Hamiltonian
                        //*********************************************************************************
                        
                        complex<double> yyderPrefac=  0.5*Stepdt*onedivdxdx*complex<double>(0.,-1./Mass);
                        
                        complex<double> ypls1valRHS =      yyderPrefac ;
                        complex<double> yvalRHS     =  -2.*yyderPrefac ;
                        complex<double> ymns1valRHS =      yyderPrefac ;

                        MatSetValues(AmatFrw,1,&pos,1,&posxyp1,&ypls1valRHS,INSERT_VALUES);
                        MatSetValues(AmatFrw,1,&pos,1,&pos   ,&yvalRHS    ,ADD_VALUES);                        
                        MatSetValues(AmatFrw,1,&pos,1,&posxym1,&ymns1valRHS,INSERT_VALUES);
   
                        //*********************************************************************************
                        //  x-contributions - 2nd order derivative to conserve positivity
                        //*********************************************************************************
                        
                        complex<double> xderPrefac= 0.5*Stepdt*(this->*Kappax)(x,y)*onedivdx;
                        
                                        xpls1valRHS =        (0.5)*xderPrefac ;
                                        xmns1valRHS =       (-0.5)*xderPrefac;

                        MatSetValues(AmatFrw,1,&pos,1,&posxp1y,&xpls1valRHS,ADD_VALUES);
                        MatSetValues(AmatFrw,1,&pos,1,&posxm1y,&xmns1valRHS,ADD_VALUES);

                        //*********************************************************************************
                        //  y-contributions - 2nd order derivative to conserve positivity
                        //*********************************************************************************
                        
                        complex<double> yderPrefac=   0.5*Stepdt*(this->*Kappay)(x,y)*onedivdx;

                                        ypls1valRHS =        (0.5)*yderPrefac ;
                                        ymns1valRHS =       (-0.5)*yderPrefac;

                        MatSetValues(AmatFrw,1,&pos,1,&posxyp1,&ypls1valRHS,ADD_VALUES);
                        MatSetValues(AmatFrw,1,&pos,1,&posxym1,&ymns1valRHS,ADD_VALUES);

                        //*********************************************************************************
                        //  mixed x y-contributions
                        //*********************************************************************************
                        
                        complex<double> xyderPrefac=  0.5*Stepdt*onedivdxdx*complex<double>((this->*Kappaxy)(x,y),0.);
                        
                        complex<double> xypp1valRHS =    0.25*xyderPrefac;
                        complex<double> xypm1valRHS =   -0.25*xyderPrefac;
                        complex<double> xymp1valRHS =   -0.25*xyderPrefac;
                        complex<double> xymm1valRHS =    0.25*xyderPrefac;
                        
                        MatSetValues(AmatFrw,1,&pos,1,&posxp1yp1,&xypp1valRHS,ADD_VALUES);
                        MatSetValues(AmatFrw,1,&pos,1,&posxm1yp1,&xypm1valRHS,ADD_VALUES);
                        MatSetValues(AmatFrw,1,&pos,1,&posxp1ym1,&xymp1valRHS,ADD_VALUES);
                        MatSetValues(AmatFrw,1,&pos,1,&posxm1ym1,&xymm1valRHS,ADD_VALUES);

                        //*********************************************************************************
                        //  x^2 - contributions - F_3
                        //*********************************************************************************
                        
                                        xxderPrefac=  0.5*Stepdt*onedivdxdx*complex<double>((this->*Kappaxx)(x,y),0.);
                        
                        complex<double> xpls2valRHS =      0.25*xxderPrefac ;
                                        xvalRHS     =  -2.*0.25*xxderPrefac ;
                        complex<double> xmns2valRHS =      0.25*xxderPrefac ;

                        MatSetValues(AmatFrw,1,&pos,1,&posxp2y,&xpls2valRHS,INSERT_VALUES);
                        MatSetValues(AmatFrw,1,&pos,1,&pos   ,&xvalRHS    ,ADD_VALUES);                        
                        MatSetValues(AmatFrw,1,&pos,1,&posxm2y,&xmns2valRHS,INSERT_VALUES);
 
                        //*********************************************************************************
                        //  y^2 - contributions  - F_3
                        //*********************************************************************************
                        
                                        yyderPrefac=  0.5*Stepdt*onedivdxdx*complex<double>((this->*Kappayy)(x,y),0.);
                        
                        complex<double> ypls2valRHS =     0.25*yyderPrefac ;
                                        yvalRHS     = -2.*0.25*yyderPrefac ;
                        complex<double> ymns2valRHS =     0.25*yyderPrefac ;

                        MatSetValues(AmatFrw,1,&pos,1,&posxyp2,&ypls2valRHS,INSERT_VALUES);
                        MatSetValues(AmatFrw,1,&pos,1,&pos   ,&yvalRHS    ,ADD_VALUES);                        
                        MatSetValues(AmatFrw,1,&pos,1,&posxym2,&ymns2valRHS,INSERT_VALUES);


                        //*********************************************************************************
                        //  diagonal terms
                        //*********************************************************************************
                        
                        double distx=Distx(x);
                        double disty=Disty(y);
                        
                        complex<double> diagvalRHS =  complex<double>(1.,0.) + complex<double>( Stepdt*0.5*(this->*Sigma)(x,y) , -Stepdt*0.5 * ( (this->*V)(distx) - (this->*V)(disty) ));
                        
                        MatSetValues(AmatFrw,1,&pos,1,&pos    ,&diagvalRHS,ADD_VALUES);
                        
                }
                        
                LocCnt++;
                 
            }
            }
        
        }

	    MatAssemblyBegin(AmatFrw,MAT_FINAL_ASSEMBLY);
	    MatAssemblyEnd(AmatFrw,MAT_FINAL_ASSEMBLY);

        MatMult(AmatFrw,rho,bvec);
        
        //*********************************************************************************
        //  LHS
        //*********************************************************************************
                 
        LocCnt=0;
        const double sgn=-1.; //-1 for CN 1 for Tst;
        
        if(BParms.WhichBEvol!=0 || InternalTimeCnt==0 || InternalTimeCnt==ReadInFromHD ){

            for(int64_t x=Axstart; x<Axend; x++){
            for(int64_t y=Aystart; y<Ayend; y++){

                        int64_t pos = PosGlobal( x, y );
                
                        int64_t posxp1y = PosGlobal( (x+1)%Ns, y );
                        int64_t posxm1y = PosGlobal( (x-1+Ns)%Ns, y );                         
                        
                        int64_t posxp2y = PosGlobal( (x+2)%Ns, y );
                        int64_t posxm2y = PosGlobal( (x-2+Ns)%Ns, y );                         
                        
                        int64_t posxyp1 = PosGlobal( x, (y+1)%Ns);
                        int64_t posxym1 = PosGlobal( x, (y-1+Ns)%Ns);                         
                         
                        int64_t posxyp2 = PosGlobal( x, (y+2)%Ns);
                        int64_t posxym2 = PosGlobal( x, (y-2+Ns)%Ns);                         
                         
                        //*****************************************
                        // NEW DERIVATIVE OPERATOR REQUIRES STEPS  
                        // IN BOTH X- AND Y-DIRECTION			   
                        //*****************************************

                        const int64_t posxp1yp1 = PosGlobal( (x+1)%Ns, (y+1)%Ns);
                        const int64_t posxm1ym1 = PosGlobal( (x-1+Ns)%Ns, (y-1+Ns)%Ns);                                                 
                        const int64_t posxp1ym1 = PosGlobal( (x+1)%Ns, (y-1+Ns)%Ns);
                        const int64_t posxm1yp1 = PosGlobal( (x-1+Ns)%Ns, (y+1)%Ns);                         

                        const int64_t posxp2yp2 = PosGlobal( (x+2)%Ns, (y+2)%Ns   );
                        const int64_t posxm2yp2 = PosGlobal( (x-2+Ns)%Ns, (y+2)%Ns);                         
                        const int64_t posxp2ym2 = PosGlobal( (x+2)%Ns, (y-2+Ns)%Ns);
                        const int64_t posxm2ym2 = PosGlobal( (x-2+Ns)%Ns, (y-2+Ns)%Ns);
                        
if(Dynamics==0){

                        //*********************************************************************************
                        //  x^2 - contributions - Hamiltonian
                        //*********************************************************************************
                        
                        complex<double> xxderPrefac=  0.5*Stepdt*onedivdxdx*complex<double>(0.,1./Mass);
                        
                        complex<double>  xpls1valRHS =      sgn*xxderPrefac ;
                        complex<double> xvalRHS     =  -2.*sgn*xxderPrefac ;
                        complex<double>  xmns1valRHS =      sgn*xxderPrefac ;

                        MatSetValues(AmatBkw,1,&pos,1,&posxp1y,&xpls1valRHS,INSERT_VALUES);
                        MatSetValues(AmatBkw,1,&pos,1,&pos   ,&xvalRHS    ,INSERT_VALUES);                        
                        MatSetValues(AmatBkw,1,&pos,1,&posxm1y,&xmns1valRHS,INSERT_VALUES);
 
                        //*********************************************************************************
                        //  y^2 - contributions - Hamiltonian
                        //*********************************************************************************
                        
                        complex<double> yyderPrefac=  0.5*Stepdt*onedivdxdx*complex<double>(0.,-1./Mass);
                        
                        complex<double> ypls1valRHS =      sgn*yyderPrefac ;
                        complex<double> yvalRHS     =  -2.*sgn*yyderPrefac ;
                        complex<double> ymns1valRHS =      sgn*yyderPrefac ;

                        MatSetValues(AmatBkw,1,&pos,1,&posxyp1,&ypls1valRHS,INSERT_VALUES);
                        MatSetValues(AmatBkw,1,&pos,1,&pos   ,&yvalRHS    ,ADD_VALUES);                        
                        MatSetValues(AmatBkw,1,&pos,1,&posxym1,&ymns1valRHS,INSERT_VALUES);
                        
                        //*********************************************************************************
                        //  diagonal terms
                        //*********************************************************************************
                        
                        double distx=Distx(x);
                        double disty=Disty(y);
                        
                        complex<double> diagvalRHS =  complex<double>(1.,0.) + complex<double>( 0. , -sgn*Stepdt*0.5 * ( (this->*V)(distx) - (this->*V)(disty) ));
                        
                        MatSetValues(AmatBkw,1,&pos,1,&pos    ,&diagvalRHS,ADD_VALUES);  
                            
    
} else if (Dynamics==1){
 
                        //*********************************************************************************
                        //  x^2 - contributions - Hamiltonian
                        //*********************************************************************************
                        
                        complex<double> xxderPrefac=  0.5*Stepdt*onedivdxdx*complex<double>(0.,1./Mass);
                        
                        complex<double> xpls1valRHS =      sgn*xxderPrefac ;
                        complex<double> xvalRHS     =  -2.*sgn*xxderPrefac ;
                        complex<double> xmns1valRHS =      sgn*xxderPrefac ;

                        MatSetValues(AmatBkw,1,&pos,1,&posxp1y,&xpls1valRHS,INSERT_VALUES);
                        MatSetValues(AmatBkw,1,&pos,1,&pos   ,&xvalRHS    ,INSERT_VALUES);                        
                        MatSetValues(AmatBkw,1,&pos,1,&posxm1y,&xmns1valRHS,INSERT_VALUES);
 
                        //*********************************************************************************
                        //  y^2 - contributions - Hamiltonian
                        //*********************************************************************************
                        
                        complex<double> yyderPrefac=  0.5*Stepdt*onedivdxdx*complex<double>(0.,-1./Mass);
                        
                        complex<double> ypls1valRHS =      sgn*yyderPrefac ;
                        complex<double> yvalRHS     =  -2.*sgn*yyderPrefac ;
                        complex<double> ymns1valRHS =      sgn*yyderPrefac ;

                        MatSetValues(AmatBkw,1,&pos,1,&posxyp1,&ypls1valRHS,INSERT_VALUES);
                        MatSetValues(AmatBkw,1,&pos,1,&pos   ,&yvalRHS    ,ADD_VALUES);                        
                        MatSetValues(AmatBkw,1,&pos,1,&posxym1,&ymns1valRHS,INSERT_VALUES);
                        
                        //*********************************************************************************
                        //  diagonal terms
                        //*********************************************************************************
                        
                        double distx=Distx(x);
                        double disty=Disty(y);
                        
                        complex<double> diagvalRHS =  complex<double>(1.,0.) + complex<double>( sgn*Stepdt*0.5*(this->*SigmaSP)(x,y) , -sgn*Stepdt*0.5 * ( (this->*V)(distx) - (this->*V)(disty) ));
                        
                        MatSetValues(AmatBkw,1,&pos,1,&pos    ,&diagvalRHS,ADD_VALUES);  
                            
    
}else if (Dynamics==2){

                        //*********************************************************************************
                        //  x^2 - contributions - Hamiltonian
                        //*********************************************************************************
                        
                        complex<double> xxderPrefac=  0.5*Stepdt*onedivdxdx*complex<double>(0.,1./Mass);
                        
                        complex<double> xpls1valRHS =      sgn*xxderPrefac ;
                        complex<double> xvalRHS     =  -2.*sgn*xxderPrefac ;
                        complex<double> xmns1valRHS =      sgn*xxderPrefac ;

                        MatSetValues(AmatBkw,1,&pos,1,&posxp1y,&xpls1valRHS,INSERT_VALUES);
                        MatSetValues(AmatBkw,1,&pos,1,&pos   ,&xvalRHS    ,INSERT_VALUES);                        
                        MatSetValues(AmatBkw,1,&pos,1,&posxm1y,&xmns1valRHS,INSERT_VALUES);
 
                        //*********************************************************************************
                        //  y^2 - contributions - Hamiltonian
                        //*********************************************************************************
                        
                        complex<double> yyderPrefac=  0.5*Stepdt*onedivdxdx*complex<double>(0.,-1./Mass);
                        
                        complex<double> ypls1valRHS =      sgn*yyderPrefac ;
                        complex<double> yvalRHS     =  -2.*sgn*yyderPrefac ;
                        complex<double> ymns1valRHS =      sgn*yyderPrefac ;

                        MatSetValues(AmatBkw,1,&pos,1,&posxyp1,&ypls1valRHS,INSERT_VALUES);
                        MatSetValues(AmatBkw,1,&pos,1,&pos   ,&yvalRHS    ,ADD_VALUES);                        
                        MatSetValues(AmatBkw,1,&pos,1,&posxym1,&ymns1valRHS,INSERT_VALUES);                        
    
                        //*********************************************************************************
                        //  x-contributions 
                        //*********************************************************************************
                        
                        complex<double> xderPrefac=   0.5*Stepdt*(this->*Kappax)(x,y)*onedivdx;

                                        xpls1valRHS =       (0.5)*sgn*xderPrefac ;
                                        xmns1valRHS =      (-0.5)*sgn*xderPrefac;

                        MatSetValues(AmatBkw,1,&pos,1,&posxp1y,&xpls1valRHS,ADD_VALUES);
                        MatSetValues(AmatBkw,1,&pos,1,&posxm1y,&xmns1valRHS,ADD_VALUES);
                                        
                        //*********************************************************************************
                        //  y-contributions 
                        //*********************************************************************************
                        
                        complex<double> yderPrefac=   0.5*Stepdt*(this->*Kappay)(x,y)*onedivdx;

                                        ypls1valRHS =        (0.5)*sgn*yderPrefac ;
                                        ymns1valRHS =       (-0.5)*sgn*yderPrefac;

                        MatSetValues(AmatBkw,1,&pos,1,&posxyp1,&ypls1valRHS,ADD_VALUES);
                        MatSetValues(AmatBkw,1,&pos,1,&posxym1,&ymns1valRHS,ADD_VALUES);
                         
                        //*********************************************************************************
                        //  mixed x y-contributions
                        //*********************************************************************************
                        
                        complex<double> xyderPrefac=  0.5*Stepdt*onedivdxdx*complex<double>((this->*Kappaxy)(x,y),0.);
                        
                        complex<double> xypp1valRHS =    0.25*sgn*xyderPrefac;
                        complex<double> xypm1valRHS =   -0.25*sgn*xyderPrefac;
                        complex<double> xymp1valRHS =   -0.25*sgn*xyderPrefac;
                        complex<double> xymm1valRHS =    0.25*sgn*xyderPrefac;
                        
                        MatSetValues(AmatBkw,1,&pos,1,&posxp1yp1,&xypp1valRHS,ADD_VALUES);
                        MatSetValues(AmatBkw,1,&pos,1,&posxm1yp1,&xypm1valRHS,ADD_VALUES);
                        MatSetValues(AmatBkw,1,&pos,1,&posxp1ym1,&xymp1valRHS,ADD_VALUES);
                        MatSetValues(AmatBkw,1,&pos,1,&posxm1ym1,&xymm1valRHS,ADD_VALUES);

                        //*********************************************************************************
                        //  x^2 - contributions - F_3
                        //*********************************************************************************
                        
                                        xxderPrefac=  0.5*Stepdt*onedivdxdx*complex<double>((this->*Kappaxx)(x,y),0.);
                        
                        complex<double> xpls2valRHS =      0.25*sgn*xxderPrefac ;
                                        xvalRHS     =  -2.*0.25*sgn*xxderPrefac ;
                        complex<double> xmns2valRHS =      0.25*sgn*xxderPrefac ;

                        MatSetValues(AmatBkw,1,&pos,1,&posxp2y,&xpls2valRHS,INSERT_VALUES);
                        MatSetValues(AmatBkw,1,&pos,1,&pos   ,&xvalRHS    ,ADD_VALUES);                        
                        MatSetValues(AmatBkw,1,&pos,1,&posxm2y,&xmns2valRHS,INSERT_VALUES);
 
                        //*********************************************************************************
                        //  y^2 - contributions  - F_3
                        //*********************************************************************************
                        
                                        yyderPrefac=  0.5*Stepdt*onedivdxdx*complex<double>((this->*Kappayy)(x,y),0.);
                        
                        complex<double> ypls2valRHS =     0.25*sgn*yyderPrefac ;
                                        yvalRHS     = -2.*0.25*sgn*yyderPrefac ;
                        complex<double> ymns2valRHS =     0.25*sgn*yyderPrefac ;

                        MatSetValues(AmatBkw,1,&pos,1,&posxyp2,&ypls2valRHS,INSERT_VALUES);
                        MatSetValues(AmatBkw,1,&pos,1,&pos   ,&yvalRHS    ,ADD_VALUES);                        
                        MatSetValues(AmatBkw,1,&pos,1,&posxym2,&ymns2valRHS,INSERT_VALUES);
                     
                        //*********************************************************************************
                        //  diagonal terms
                        //*********************************************************************************
                        
                        double distx=Distx(x);
                        double disty=Disty(y);
                        
                        complex<double> diagvalRHS =  complex<double>(1.,0.) + complex<double>( sgn*Stepdt*0.5*(this->*Sigma)(x,y) , -sgn*Stepdt*0.5 * ( (this->*V)(distx) - (this->*V)(disty) ));
                        
                        MatSetValues(AmatBkw,1,&pos,1,&pos    ,&diagvalRHS,ADD_VALUES);  
                         
                }                        
                        
                LocCnt++;
                 
            }
            }

        }

	    MatAssemblyBegin(AmatBkw,MAT_FINAL_ASSEMBLY);
	    MatAssemblyEnd(AmatBkw,MAT_FINAL_ASSEMBLY);
        
        //*********************************************************************************
        //  Solve the linear system iteratively 
        //*********************************************************************************
        
        ierr = KSPSolve(ksp,bvec,rho);

    }

    void DensityMatrix::OutputRhoMidSliceRsToHD_ASCII(char* Filename)
    {
        MPI_Status status;
        MPI_Barrier(PETSC_COMM_WORLD);
        
        const complex<double>* RhoLocal;
        VecGetArrayRead(rho,&RhoLocal);
                
        int Semaphore=NodeRank;
        
        FILE* OutputRho;
        
        if(NodeRank==0) { OutputRho=fopen(Filename,"w"); fclose(OutputRho); }
        
        if(NodeRank!=0)  MPI_Recv(&Semaphore, 1, MPI_INT, NodeRank-1, MPI_ANY_TAG , PETSC_COMM_WORLD, &status);
        
        int64_t LocCnt=0;
        OutputRho=fopen(Filename,"a");
        
        for(int64_t x=xstart; x<xend; x++){
        for(int64_t y=ystart; y<yend; y++){

            double R = ( x + y - Ns )*ds/2.;
            double r = ( x - y )*ds/2.;
            
            fprintf(OutputRho, "%e %e %e %e\n", R,r, RhoLocal[ LocCnt ].real(), RhoLocal[ LocCnt ].imag() );
            
            LocCnt++;
        }
        }
        fclose(OutputRho);
        
        if(NodeRank<NodeNum-1)  MPI_Send(&Semaphore, 1, MPI_INT, NodeRank+1, 1 , PETSC_COMM_WORLD);
        
        VecRestoreArrayRead(rho,&RhoLocal); 
        
        MPI_Barrier(PETSC_COMM_WORLD);
 
    }      
    
    void DensityMatrix::OutputRhoToHD(char* Filename, int64_t curtime)
    {

        MPI_Barrier(PETSC_COMM_WORLD);
        
        const complex<double>* RhoLocal;
        VecGetArrayRead(rho,&RhoLocal);
        
        char FilenamePerNode[512];
        
        sprintf(FilenamePerNode, "%s%ld", Filename,NodeRank);
        
        std::fstream OutputRho;
        OutputRho.open(FilenamePerNode, std::ios::out|std::ios::binary);
        
        OutputRho.write ((char*) &NodeRank, sizeof(int64_t));			
        OutputRho.write ((char*) &curtime, sizeof(int64_t));
        
        int64_t CurSize=(AIend-AIstart);

        OutputRho.write ((char*) &CurSize, sizeof(int64_t));
        
        OutputRho.write ((char*) RhoLocal, sizeof(complex<double>)*(AIend-AIstart));			
            
        OutputRho.close();
        
        VecRestoreArrayRead(rho,&RhoLocal); 
        
        MPI_Barrier(PETSC_COMM_WORLD);
        
    }
    
    
    void DensityMatrix::ReadInRhoFromHD(char* Filename, int64_t *curtime)
    {
        
        MPI_Barrier(PETSC_COMM_WORLD);
        
        complex<double> *RhoLocal;
        VecGetArray(rho,&RhoLocal);
            
        char FilenamePerNode[512];
        
        sprintf(FilenamePerNode, "%s%ld", Filename,NodeRank);
        
        std::fstream InputRho;
        InputRho.open(FilenamePerNode, std::ios::in|std::ios::binary);
        
        int64_t HDRank;
        
        InputRho.read ((char*) &HDRank, sizeof(int64_t));	
        
        fprintf(stderr, "File on HD from node: %ld - read by current node: %ld\n",HDRank,NodeRank);
        
        
        InputRho.read ((char*) curtime, sizeof(int64_t));
        
        int64_t CurSize;
        InputRho.read ((char*) CurSize, sizeof(int64_t));
        
        if(CurSize!=(AIend-AIstart)){ fprintf(stderr, "Node:%ld -- Size on HD: %ld does NOT match current size: %ld\n",NodeRank,CurSize,(AIend-AIstart)); exit(-1);}
        
        InputRho.read ((char*) RhoLocal, sizeof(complex<double>)*(AIend-AIstart));
            
        InputRho.close();
        
        VecRestoreArray(rho,&RhoLocal);
        
        MPI_Barrier(PETSC_COMM_WORLD);

        ReadInFromHD=*curtime;
        
    }
    
    void DensityMatrix::OutputInteractionsAndTemp(char* Filename)
    {            
            FILE* OutputIntAndT;
            
            OutputIntAndT=fopen(Filename,"w");
                    
                for(int64_t x=0; x<Ns; x++){
                    
                    double dist = Distx(x);
                    
                    double Temp=(this->*Temperature)();
                    
                    double StatePot=0.;
                    
                    fprintf(OutputIntAndT, "%e %e %e %e %e %e\n", dist, (this->*V)(dist), (this->*ImV)(dist), Temp, (this->*CorrelationLength)(Temp), StatePot);
                    
                }
        
            fclose(OutputIntAndT);
      
    }  

//*********************************************************************************
//   MAIN PROGRAM BODY
//*********************************************************************************

static char help[] = "Solves 1d master equation with CN.\n\n";

int main(int argc,char *argv[])
{

    //*********************************************************************************
    //  PETSC & MPI INITIALIZATION
    //*********************************************************************************
    
        int MPIMyRank;
        int MPINumProcs;
        
        PetscErrorCode ierr;        
        int noargs =1;
        SlepcInitialize(&noargs,&argv,(char*)0,(char*)0);
        
        ierr = MPI_Comm_size(PETSC_COMM_WORLD,&MPINumProcs);CHKERRQ(ierr);
        ierr = MPI_Comm_rank(PETSC_COMM_WORLD,&MPIMyRank);CHKERRQ(ierr);
        
        fprintf(stderr, "Num: %ld, Rank:%ld\n",MPINumProcs,MPIMyRank);
        
        CHKERRQ(ierr);   
        
        //*********************************************************************************
        //  BEGIN WITH THE MAIN PROGRAM PART
        //*********************************************************************************
    
        time_t start,current;
        time (&start);
        int64_t SaveCnt=1;
        
	//*********************************************************************************
	//   INITIALIZE THE PROGRAM AND READ COMMAND LINE OPTIONS
	//*********************************************************************************
	
	//*********************************************************************************
	//  COMMAND LINE ALLOWS TO SPECIFY THE FOLLOWING
	//*********************************************************************************

        if(argc!=37) { if(MPIMyRank==0) fprintf(stderr, "Provided number of parameters %3d not same as expected. Please use run script.\n",argc); MPI_Barrier(PETSC_COMM_WORLD); MPI_Finalize(); exit(-1);}

            //*****************************************************************************
            //  THREADS, GEOMETRY, INTERACTIONS, INITIAL CONDITIONS, TEMPERATURE, ...  
            //*****************************************************************************

                int64_t Threads=atoi(argv[1]);
                
                Threads=0;
                
                int64_t BoxSize=atoi(argv[2]);
                
                double ds=atof(argv[3]);
                double dt=atof(argv[4]);            
            
                double Mass=atof(argv[5]);
                double DistReg=1./Mass;
                
            InteractionParms MyIParms;
            
                int64_t WhichIntReV=atoi(argv[6]);
                int64_t WhichIntD=atoi(argv[7]);
            
                MyIParms.alpha=atof(argv[8]);
                MyIParms.sigma=atof(argv[9]);
                MyIParms.c=atof(argv[10]);
                
                MyIParms.gamma=atof(argv[11]);
                MyIParms.mD=atof(argv[12]);
                
                MyIParms.DistReg=DistReg;
        
            BackgroundParms MyBParms;
            
                int64_t WhichBEvol=atoi(argv[13]);
            
                MyBParms.WhichBEvol=WhichBEvol;
                MyBParms.Tfinal=atof(argv[33]);
                MyBParms.Tinit=atof(argv[14]);
                MyBParms.lcorr=atof(argv[15]);
                
                MyBParms.lcorrPropTFac=atof(argv[34]);
                MyBParms.mDPropTFac=atof(argv[35]);
                MyBParms.gammaPropTFac=atof(argv[36]);
            
            StatesParms MySParms;

                MySParms.alpha=atof(argv[16]);
                MySParms.sigma=atof(argv[17]);
                MySParms.c=atof(argv[18]);
                MySParms.mD=atof(argv[31]);
                MySParms.DistReg=DistReg;
                sprintf(MySParms.BasisFuncsHD,"%s",argv[19]);
            
                int64_t WhatStates=atoi(argv[20]);
                int64_t HowManyStates=atoi(argv[21]);

            InitCondParms MyICParms;
            
                int64_t WhatIC=atoi(argv[22]);
            
                MyICParms.sigmas=atof(argv[23]);
                MyICParms.sigmaR=atof(argv[24]);
                
                MyICParms.SelectInitState=atoi(argv[25]);
            
                MyICParms.StartStep=0;
                
                sprintf(MyICParms.LocationOfSavedState,"%s",argv[27]);
                
                
            int64_t OutputEvery=atoi(argv[26]);
            
            char OutputDir[256];
            
            sprintf(OutputDir,"%s",argv[27]);
        
            int64_t EvolSteps=atof(argv[28]);

            VERBOSE=atoi(argv[29]);

            int64_t WhichDynamics=atof(argv[30]);
            
            if(MPIMyRank==0){
            char cmd[256];
            sprintf(cmd, "mkdir %s", OutputDir);
            system (cmd);
            }
            
            double SaveAfterHowManyHours=atof(argv[32]);
            
            int64_t SaveEvery = abs(SaveAfterHowManyHours);
            
            MPI_Barrier(PETSC_COMM_WORLD);
            
	//*********************************************************************************
	//  ALLOCATE THE NECESSARY OBJECTS AND SET THE PARAMETERS
	//*********************************************************************************
                        
            DensityMatrix Rho;
            
            if(MPIMyRank==0){
            fprintf(stderr, "******************************************************\n");
            fprintf(stderr, "*** Setup of the Denisty Matrix                    ***\n");
            fprintf(stderr, "*** Allocation of memory                           ***\n");
            }
            MPI_Barrier(PETSC_COMM_WORLD);
            
            Rho.InitializeGeometry(BoxSize,ds,dt,Mass,0, MPINumProcs, MPIMyRank, 0, WhichDynamics);
            MPI_Barrier(PETSC_COMM_WORLD);
                        
            if(MPIMyRank==0){
            fprintf(stderr, "*** Select which Interaction                       ***\n");
            }
            Rho.SetupInteractions(WhichIntReV,WhichIntD,MyIParms);
            MPI_Barrier(PETSC_COMM_WORLD);
            
            if(MPIMyRank==0){
            fprintf(stderr, "*** Select which Background evolution              ***\n");
            }
            Rho.SetupBackgroundEvol(WhichBEvol, MyBParms);
            MPI_Barrier(PETSC_COMM_WORLD);
            
            if(MPIMyRank==0){
            fprintf(stderr, "*** Setup the basis states                         ***\n");
            }
            Rho.SetupBasisStates(WhatStates, HowManyStates,MySParms);
            MPI_Barrier(PETSC_COMM_WORLD);

            if(MPIMyRank==0){
            fprintf(stderr, "*** Set initial conditions for Rho                 ***\n");
            }
            Rho.InitializeDensityMatrix(WhatIC, MyICParms);
            MPI_Barrier(PETSC_COMM_WORLD);
                                    
            if(MPIMyRank==0){
            fprintf(stderr, "******************************************************\n\n");
            }
            
            MPI_Barrier(PETSC_COMM_WORLD);
            
	//*********************************************************************************
	//  OUTPUT THE INITIAL PARAMETERS TO SCREEN
	//********************************************************************************* 
            
            if(MPIMyRank==0){
                
            fprintf(stderr, "******************************************************\n");
            fprintf(stderr, "*** Evolution of the 1-dimensional density matrix  ***\n");
            
            if (WhichDynamics==0){
            fprintf(stderr, "*** Only with the ReV NO fluct/diss  2nd order CN  ***\n");                
            }else if (WhichDynamics==1) {
            fprintf(stderr, "*** Stochastic Potential approx      2nd order CN  ***\n");    
            }else if (WhichDynamics==2) {
            fprintf(stderr, "*** Full Lindblad Equation           2nd order CN  ***\n");
            }else{
            fprintf(stderr, "*** Please choose valid Dynamics 0,1,2             ***\n"); exit(-1);   
            }
            
            fprintf(stderr, "***   with %3d processes via MPI                   ***\n", MPINumProcs);
            fprintf(stderr, "***   Evolving %7ld steps                       ***\n", EvolSteps);
            fprintf(stderr, "***   Ns=%3ld            ds=%.2e                ***\n", Rho.ReturnNs(), Rho.Returnds());
            fprintf(stderr, "***   dt=%.2e Mass=%.2e                    ***\n", Rho.Returndt(),Rho.ReturnMass());
            fprintf(stderr, "***   WhichInteraction: V:%3ld D:%3ld                ***\n", WhichIntReV,WhichIntD);
            fprintf(stderr, "***   WhichBackground: %3ld                         ***\n", WhichBEvol);
            fprintf(stderr, "***   WhichStates: %3ld Howmany: %3ld                ***\n", WhatStates, HowManyStates);
            fprintf(stderr, "***   WhichInitialCond: %3ld                        ***\n", WhatIC);
            fprintf(stderr, "***   Save State every: %3ld steps                  ***\n", SaveEvery);
            fprintf(stderr, "******************************************************\n\n");
            
            }
            
            MPI_Barrier(PETSC_COMM_WORLD);
            
	//*********************************************************************************
	//  OUTPUT THE STATES TO HD
	//*********************************************************************************
           
            char StatesFname[256], RhoFname[256], IntFname[256];
            FILE* TraceOutput;
            FILE* BackgroundOutput;
            FILE* OutputRho;
            FILE* EnergyOutput;
            double TraceAbs, GlobalTraceAbs;
            complex<double> Trace, GlobalTrace, Energy, GlobalEnergy;

            complex<double> *RhoStateBasis,*RhoStateBasisGlobal;
        
                             RhoStateBasis       = new complex<double> [HowManyStates*HowManyStates];
                             RhoStateBasisGlobal = new complex<double> [HowManyStates*HowManyStates];
            
                        
            if(Rho.ReturnStartTimeStep()==0){
        
                    Trace =  Rho.ReturnTrace();
                    MPI_Reduce(&Trace, &GlobalTrace, 1, MPI_DOUBLE_COMPLEX, MPI_SUM, 0,PETSC_COMM_WORLD);
                
                    TraceAbs =  Rho.ReturnTraceAbs();
                    MPI_Reduce(&TraceAbs, &GlobalTraceAbs, 1, MPI_DOUBLE, MPI_SUM, 0,PETSC_COMM_WORLD);
                
                    GlobalEnergy =  Rho.ReturnEnergy();

                    Rho.DensityMatrixInStateBasis(RhoStateBasis);        
                    MPI_Reduce(RhoStateBasis, RhoStateBasisGlobal, HowManyStates*HowManyStates, MPI_DOUBLE_COMPLEX, MPI_SUM, 0,PETSC_COMM_WORLD);
                    

                    if(MPIMyRank==0){
                        
                        sprintf(RhoFname, "%sBackgroundEvol.dat",OutputDir);
                        BackgroundOutput=fopen(RhoFname,"w");
                        fprintf(BackgroundOutput, "%d %e %.20e %.20e\n",0,0.,Rho.ReturnCurrentTemp(),Rho.ReturnCurrentCorrelationLength());
                        fclose(BackgroundOutput);
                            
                        sprintf(RhoFname, "%sTraceEvol.dat",OutputDir);
                        TraceOutput=fopen(RhoFname,"w");
                        fprintf(TraceOutput, "%d %e %.20e %.20e %.20e \n",0,0.,GlobalTrace.real(),GlobalTrace.imag(),GlobalTraceAbs);
                        fclose(TraceOutput);                
              
                        sprintf(RhoFname, "%sEnergyEvol.dat",OutputDir);
                        EnergyOutput=fopen(RhoFname,"w");
                        fprintf(EnergyOutput, "%d %e %.20e %.20e\n",0,0.,GlobalEnergy.real(),GlobalEnergy.imag());
                        fclose(EnergyOutput);                
              
                        sprintf(StatesFname, "%sBasisStates.dat",OutputDir);
                        sprintf(IntFname, "%sInteractionsAndTemp.dat",OutputDir);

                        Rho.OutputInteractionsAndTemp(IntFname);
                                          
                        sprintf(RhoFname, "%sDensityMatrixStateBasisStep_0.dat",OutputDir);
                        OutputRho=fopen(RhoFname,"w");
                        
                        if(OutputRho==0){ fprintf(stderr, "PROBLEM OPENING %s\n", RhoFname); exit(-1);}
                        
                        double OverallProb=0.;
                        
                        for(int64_t scnt1=0; scnt1<HowManyStates; scnt1++){  
                            for(int64_t scnt2=0; scnt2<HowManyStates; scnt2++){  

                            fprintf(OutputRho, "%ld %ld %.20e %.20e\n", scnt1, scnt2, RhoStateBasisGlobal[scnt1*HowManyStates+scnt2].real(), RhoStateBasisGlobal[scnt1*HowManyStates+scnt2].imag());
                                
                        } OverallProb+=RhoStateBasisGlobal[scnt1*HowManyStates+scnt1].real(); }
                            
                        fprintf(OutputRho, "# Overall Prob: %e\n",  OverallProb);
                            
                        fclose(OutputRho);
                        
                        sprintf(RhoFname, "%sStatesEvol.dat",OutputDir);
                        BackgroundOutput=fopen(RhoFname,"w");
                        fprintf(BackgroundOutput, "%d %e ",0,0.);

                        for(int64_t scnt1=0; scnt1<HowManyStates; scnt1++){
                        fprintf(BackgroundOutput, "%.20e ",RhoStateBasisGlobal[scnt1*HowManyStates+scnt1].real());
                        }
                    
                        fprintf(BackgroundOutput, "\n");
                        fclose(BackgroundOutput);
                        
                    }
    
                    sprintf(RhoFname, "%sDensityMatrixPosBasisMidSliceRsStep_%d.dat",OutputDir,0);
                    Rho.OutputRhoMidSliceRsToHD_ASCII(RhoFname);

                    Rho.OutputEigenvalsToHD(0,OutputDir);
                   
       }
          
	//*********************************************************************************
	//  ALLOCATE THE NECESSARY OBJECTS AND SET THE PARAMETERS
	//*********************************************************************************

            if (MPIMyRank==0) fprintf(stderr, "******************************************************\n");
    
            for(int64_t tstep=Rho.ReturnStartTimeStep(); tstep<EvolSteps;  tstep++){

                MPI_Barrier(PETSC_COMM_WORLD);

                if (VERBOSE==1&&MPIMyRank==0) fprintf(stderr, "***                                                ***\n");
                if (VERBOSE==1&&MPIMyRank==0) fprintf(stderr, "***  Evolution Step: %8ld - t=%.3e        ***\n", tstep,tstep*dt);
                if (VERBOSE==1&&MPIMyRank==0) fprintf(stderr, "***                                                ***\n");

                Rho.ApplyFullEvolutionOperatorUsingPETSC( Rho.Returndt());

                //*********************************************************************************
                //  END OF DYNAMICS
                //*********************************************************************************
                 
                MPI_Barrier(PETSC_COMM_WORLD);
                
                Rho.EvolStepCompleted();

                //*********************************************************************************
                // OUTPUT OF TRACE AND BASIS STATE PROJECTION
                //*********************************************************************************
        
                if((tstep+1)%OutputEvery==0){
                    
                    Trace =  Rho.ReturnTrace();
                    MPI_Reduce(&Trace, &GlobalTrace, 1, MPI_DOUBLE_COMPLEX, MPI_SUM, 0,PETSC_COMM_WORLD);
                
                    TraceAbs =  Rho.ReturnTraceAbs();
                    MPI_Reduce(&TraceAbs, &GlobalTraceAbs, 1, MPI_DOUBLE, MPI_SUM, 0,PETSC_COMM_WORLD);
                    
                    GlobalEnergy =  Rho.ReturnEnergy();
                
                    Rho.DensityMatrixInStateBasis(RhoStateBasis);
                    MPI_Reduce(RhoStateBasis, RhoStateBasisGlobal, HowManyStates*HowManyStates, MPI_DOUBLE_COMPLEX, MPI_SUM, 0,PETSC_COMM_WORLD);
                    
                    if (VERBOSE==1&&MPIMyRank==0) fprintf(stderr, "***  t=%8ld Trace value:  %.10e %.10e ***\n", tstep,GlobalTrace.real(),GlobalTrace.imag());
                    
                    if(MPIMyRank==0){
                        
                        sprintf(RhoFname, "%sTraceEvol.dat",OutputDir);
                        TraceOutput=fopen(RhoFname,"a");
                        fprintf(TraceOutput, "%ld %e %.20e %.20e %.20e\n",tstep+1,(tstep+1)*dt,GlobalTrace.real(),GlobalTrace.imag(),GlobalTraceAbs);
                        fclose(TraceOutput);  
                        
                        sprintf(RhoFname, "%sEnergyEvol.dat",OutputDir);
                        EnergyOutput=fopen(RhoFname,"a");
                        fprintf(EnergyOutput, "%ld %e %.20e %.20e\n",tstep+1,(tstep+1)*dt,GlobalEnergy.real(),GlobalEnergy.imag());
                        fclose(EnergyOutput);  
                    
                        sprintf(RhoFname, "%sBackgroundEvol.dat",OutputDir);
                        BackgroundOutput=fopen(RhoFname,"a");
                        fprintf(BackgroundOutput, "%ld %e %.20e %.20e\n",tstep+1,(tstep+1)*dt,Rho.ReturnCurrentTemp(),Rho.ReturnCurrentCorrelationLength());
                        fclose(BackgroundOutput);
           
                        sprintf(RhoFname, "%sDensityMatrixStateBasisStep_%ld.dat",OutputDir,tstep+1);
                        OutputRho=fopen(RhoFname,"w");

                        double OverallProb=0.;
                        
                        for(int64_t scnt1=0; scnt1<HowManyStates; scnt1++){  
                            for(int64_t scnt2=0; scnt2<HowManyStates; scnt2++){  

                            fprintf(OutputRho, "%ld %ld %.20e %.20e\n", scnt1, scnt2, RhoStateBasisGlobal[scnt1*HowManyStates+scnt2].real(), RhoStateBasisGlobal[scnt1*HowManyStates+scnt2].imag());
                            
                            
                                
                        } OverallProb+=RhoStateBasisGlobal[scnt1*HowManyStates+scnt1].real(); }
                            
                        fprintf(OutputRho, "# Overall Prob: %e\n",  OverallProb);
                            
                        fclose(OutputRho);

                        sprintf(RhoFname, "%sStatesEvol.dat",OutputDir);
                        BackgroundOutput=fopen(RhoFname,"a");
                        fprintf(BackgroundOutput, "%d %e ",tstep+1,(tstep+1)*dt);
                        
                        for(int64_t scnt1=0; scnt1<HowManyStates; scnt1++){
                        fprintf(BackgroundOutput, "%.20e ",RhoStateBasisGlobal[scnt1*HowManyStates+scnt1].real());
                        }
                        
                        fprintf(BackgroundOutput, "\n");
                        fclose(BackgroundOutput);
                        
                    }
              
                    if((tstep+1)%(OutputEvery*10)==0){
                    sprintf(RhoFname, "%sDensityMatrixPosBasisMidSliceRsStep_%ld.dat",OutputDir,tstep+1);
                    Rho.OutputRhoMidSliceRsToHD_ASCII(RhoFname);
                    }
                    
                    Rho.OutputEigenvalsToHD(tstep+1,OutputDir);
                    
                }
                
                //*********************************************************************************
                //  OUTPUT OF FULL STATE AFTER CERTAIN NUMBER OF EVOLUTION STEPS
                //*********************************************************************************
                    
                if((tstep+1)%SaveEvery==0&&tstep>Rho.ReturnStartTimeStep()){
                    
                    if(MPIMyRank==0) fprintf(stderr, "OUTPUTTING RHO TO HD AT tstep:%ld\n", tstep);
                  
                    char fname[256];
                    sprintf(fname,"%sCurrentRho.dat",OutputDir);
                    int64_t curtime=tstep+1;
                    
                    Rho.OutputRhoToHD(fname,curtime);
                    
                    if(MPIMyRank==0) fprintf(stderr, "Shutting down so that not time wasted for restart\n");
                    
                    MPI_Barrier(PETSC_COMM_WORLD);
                    MPI_Finalize(); 
                    exit(0);
                                        
                }
                
        
            }
            
            if (MPIMyRank==0) fprintf(stderr, "******************************************************\n\n");
   
            
        MPI_Finalize();
        
        return(0);
}

  


